////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// XPS Printer Driver SDK: card completion for non-interactive mode jobs. c++ sample.
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <bidispl.h>
#include <WinSpool.h>
#include <iostream>
#include <sstream>
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

void usage(CString longExeName)
{
    CPath shortExeName(longExeName);
    shortExeName.StripPath();
    cout << endl;
    cout << CT2A(shortExeName) << " demonstrates the card completion functionality " << endl;
    cout << "for print jobs using Windows printing." << endl << endl;
    cout << CT2A(shortExeName) << " -n <printername>" << endl << endl;
    cout << "options:" << endl;
    cout << "  " << "-n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
    exit(0);
}

void CancelCard(CComPtr <IBidiSpl>& bidiSpl, const util::PrinterStatusValues& printerStatusValues)
{
    stringstream exceptionText;
    CComPtr <IBidiRequest> bidiRequest;
    HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
    if (FAILED(hr)) {
        exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
        throw runtime_error(exceptionText.str());
    }

    // create the XML and convert to bytes for the upcoming SetInputData() call.
    // NOTE: It is IMPORTANT to set the errorcode member; it clears that
    // particular error in the printer.
    CString actionXML = util::FormatPrinterActionXML(
        dxp01sdk::Cancel,
        printerStatusValues._printerJobID,
        printerStatusValues._errorCode);

    const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

    // IBidiSpl compatible memory:
    CComAllocator XMLBytesAllocator;
    PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
    DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

    // copy the XML string...without the trailing NULL:
    ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

    hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
    if (FAILED(hr)) {
        exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
        throw runtime_error(exceptionText.str());
    }

    hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
    if (FAILED(hr)) {
        exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
        throw runtime_error(exceptionText.str());
    }

    hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
    if (FAILED(hr)){
        exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
        throw runtime_error(exceptionText.str());
    }
}

////////////////////////////////////////////////////////////////////////////////
// PollForJobCompletion()
//
// repeatedly check job status until the job is complete.
////////////////////////////////////////////////////////////////////////////////
void PollForJobCompletion(
    CComPtr <IBidiSpl>&  bidiSpl,
    const int            printerJobID,
    const int            intervalMilliseconds)
{
    for (; true ;) {
        const CString jobStatusXML = util::GetJobStatusXML(bidiSpl, printerJobID);
        const util::JobStatusValues jobStatusValues = util::ParseJobStatusXML(jobStatusXML);
        cout << "Printer job id " << printerJobID << " " << CW2A(jobStatusValues._jobState) << endl;
        if (dxp01sdk::JOB_ACTIVE != jobStatusValues._jobState) {
            return;
        }
        ::Sleep(intervalMilliseconds);
    }
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
    stringstream exceptionText;
    CString printerName;
    int c(0);

    while ((c = getopt(argc, argv, _T("n:"))) != EOF) {
        switch (c) {
        case L'n':
            printerName = optarg; break;
        default:
            usage(argv[0]);
        }
    }

    if (printerName.IsEmpty()) {
        usage(argv[0]);
    }

    CComPtr <IBidiSpl> bidiSpl;
    try {
        HRESULT hr = ::CoInitialize(NULL);
        if (FAILED(hr)) {
            exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(exceptionText.str());
        }

        // create an IBidiSpl COM smartpointer instance and bind it to the printer
        // name.
        hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
        if (FAILED(hr)) {
            exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(exceptionText.str());
        }

        hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
        if (FAILED(hr)) {
            exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(printerName) << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(exceptionText.str());
        }

        const CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
        const CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
        cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

        // is the printer is in the ready state?
        const CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
        util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);
        if ("Ready" != printerOptionsValues._printerStatus) {
            exceptionText << "printer " << CT2A(printerName) << " is not ready.";
            throw runtime_error(exceptionText.str());
        }

        // get the printer handle for printing:
        HANDLE printerHandle(0);
        BOOL brc = ::OpenPrinter(printerName.GetBuffer(), &printerHandle, NULL);
        if (!brc) {
            const DWORD lastErr = ::GetLastError();
            exceptionText << "OpenPrinter(): " << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
            throw runtime_error(exceptionText.str());
        }

        // do some windows printing and get the windowsJobID:
        DWORD windowsJobID(0);
        const bool doTwoSided = printerOptionsValues._optionDuplex == "Auto";
        util::PrintTextAndGraphics(printerName, printerHandle, windowsJobID, doTwoSided);

        // Important: wait until driver gets all the print data:
        util::WaitUntilJobSpooled(printerHandle, windowsJobID); 
        ::ClosePrinter(printerHandle);

        // Get the printer job id to check for card completion. 
        // We have to retrieve the printer job id as job was not started using
        // the StartJob method - which returns the printer job id.
        CString printerStatusXML = util::GetPrinterStatusXML(bidiSpl);
        util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode) {
            // the Job failed; send a Cancel(); 
            CancelCard(bidiSpl, printerStatusValues);
            cout << "Job failed. Cancel the card. Windows job id:"
                << printerStatusValues._windowsJobID 
                << " , Printer job id:"  
                << printerStatusValues._printerJobID << endl;
        }

        // Important: make sure that windows jobID matches the printer jobID:
        if (windowsJobID == (DWORD)printerStatusValues._windowsJobID) {
            PollForJobCompletion(bidiSpl, printerStatusValues._printerJobID, 5000);
        }
        else
        {
            exceptionText << "Failed to get valid printer jobId for windows jobid: " << windowsJobID;
            throw runtime_error(exceptionText.str());
        }
    }
    catch (runtime_error& e) {
        cerr << e.what() << endl;
    }

    if (bidiSpl) {
        bidiSpl->UnbindDevice();
        bidiSpl = NULL;
    }

    ::CoUninitialize();

    return 0;
}
