////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// common 'util' methods are used throughout the xps driver sdk samples
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <iostream>
#include <sstream>
#include <atlenc.h>
#include <atltime.h>
#include "DXP01SDK.H"
#include "util.h"

using namespace std;

////////////////////////////////////////////////////////////////////////////////
// util::Win32ErrorString()
// for the given Win32 error code, return a unicode string
////////////////////////////////////////////////////////////////////////////////
CString util::Win32ErrorString(const long errorCode)
{
   CString errorString;
   const DWORD bufsize(512);
   ::FormatMessage(
      FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_FROM_SYSTEM,
      NULL,
      errorCode,
      MAKELANGID(LANG_NEUTRAL, SUBLANG_NEUTRAL),
      errorString.GetBuffer(bufsize),
      bufsize,
      NULL);
   errorString.ReleaseBuffer();
   errorString.Trim();
   return errorString;
}

////////////////////////////////////////////////////////////////////////////////
// util::EchoCurrentTime()
////////////////////////////////////////////////////////////////////////////////
void util::EchoCurrentTime(CString msg)
{
   const CTime current_time(CTime::GetCurrentTime());
   const CString current_time_formatted(current_time.Format(TEXT("%Y/%m/%d %H:%M:%S")));
   cout << CT2A(current_time_formatted) << " " << CT2A(msg) << endl;
}

////////////////////////////////////////////////////////////////////////////////
// util::StringFromXml()
//
// for the given XML document, return a string value using the given xpath
// expression.
//
// if the string is not found, send a msg to the debug output, and return an
// empty string.
////////////////////////////////////////////////////////////////////////////////
CString util::StringFromXml(CComPtr <IXMLDOMDocument2>& doc, const PWCHAR xpathExpression)
{
   CString val;
   CComPtr <IXMLDOMNode> node;
   stringstream utilExceptionText;

   HRESULT hr = doc->selectSingleNode(xpathExpression, &node);
   if (FAILED(hr) || S_FALSE == hr) {
      utilExceptionText << "doc->selectSingleNode() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr)) << endl;
      AtlTrace("%s\n", utilExceptionText.str().c_str());
      return val;
   }

   CComBSTR nodeText;
   hr = node->get_text(&nodeText);
   if (FAILED(hr)) {
      utilExceptionText << "node->get_text() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      AtlTrace("%s\n", utilExceptionText.str().c_str());
      return val;
   }

   val = nodeText;

   return val;
}

////////////////////////////////////////////////////////////////////////////////
// util::IntFromXml()
// return a numeric value from the xml doc using the given xpath expression
// return zero if not found.
////////////////////////////////////////////////////////////////////////////////
int util::IntFromXml(CComPtr <IXMLDOMDocument2>& doc, const PWCHAR xpathExpression)
{
   int val(0);
   CComPtr <IXMLDOMNode> node;
   stringstream utilExceptionText;

   HRESULT hr = doc->selectSingleNode(xpathExpression, &node);
   if (FAILED(hr) || S_FALSE == hr) {
      utilExceptionText << "doc->selectSingleNode() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr)) << endl;
      AtlTrace("%s\n", utilExceptionText.str().c_str());
      return val;
   }

   CComBSTR nodeText;
   hr = node->get_text(&nodeText);
   if (FAILED(hr)) {
      utilExceptionText << "node->get_text() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " "
         << CT2A(util::Win32ErrorString(hr))
         << endl;
      AtlTrace("%s", utilExceptionText.str());
      return val;
   }

   val = ::_wtoi(nodeText);

   return val;
}

////////////////////////////////////////////////////////////////////////////////
// util::BoolFromXml()
// return true or false for the given xpath expression.
// if not found...return false.
////////////////////////////////////////////////////////////////////////////////
bool util::BoolFromXml(CComPtr <IXMLDOMDocument2>& doc, const PWCHAR xpathExpression)
{
   bool val(false);
   CComPtr <IXMLDOMNode> node;
   stringstream utilExceptionText;

   HRESULT hr = doc->selectSingleNode(xpathExpression, &node);
   if (FAILED(hr) || S_FALSE == hr) {
      utilExceptionText << "doc->selectSingleNode() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr)) << endl;
      AtlTrace("%s\n", utilExceptionText.str().c_str());
      return val;
   }

   CComBSTR nodeText;
   hr = node->get_text(&nodeText);
   if (FAILED(hr)) {
      utilExceptionText << "node->get_text() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " "
         << CT2A(util::Win32ErrorString(hr))
         << endl;
      AtlTrace("%s", utilExceptionText.str());
      return val;
   }

   val = ::_wtoi(nodeText) ? true : false;

   return val;
}

////////////////////////////////////////////////////////////////////////////////
// util::UlongFromXml()
// return a numeric value from the xml doc using the given xpath expression
// return zero if not found.
////////////////////////////////////////////////////////////////////////////////
ULONG util::UlongFromXml(CComPtr <IXMLDOMDocument2>& doc, const PWCHAR xpathExpression)
{
   ULONG val(0);
   CComPtr <IXMLDOMNode> node;
   stringstream utilExceptionText;

   HRESULT hr = doc->selectSingleNode(xpathExpression, &node);
   if (FAILED(hr) || S_FALSE == hr) {
      utilExceptionText << "doc->selectSingleNode() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr)) << endl;
      AtlTrace("%s\n", utilExceptionText.str().c_str());
      return val;
   }

   CComBSTR nodeText;
   hr = node->get_text(&nodeText);
   if (FAILED(hr)) {
      utilExceptionText << "node->get_text() " << CW2A(xpathExpression)
         << " 0x" << hex << hr << " "
         << CT2A(util::Win32ErrorString(hr))
         << endl;
      AtlTrace("%s", utilExceptionText.str());
      return val;
   }

   val = ::_wtol(nodeText);

   return val;
}

////////////////////////////////////////////////////////////////////////////////
// util::GetPrinterStatusXML()
//
// Most commands are followed with a call to this function. It returns xml
// markup that contains the most recent state of the printer.
//
// This version of
// GetPrinterStatusXML() is uses the most recently issued IBidiRequest.
////////////////////////////////////////////////////////////////////////////////
CString util::GetPrinterStatusXML(CComPtr <IBidiRequest>& bidiRequest)
{
   stringstream utilExceptionText;
   LPWSTR schema(NULL);
   PBYTE dataPointer(NULL);
   DWORD dataType = 0;
   ULONG dataSizeInBytes = 0;
   HRESULT hr = bidiRequest->GetOutputData(
      0,
      &schema,
      &dataType,
      &dataPointer,
      &dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "GetOutputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR printerStatusString;
   hr = printerStatusString.AppendBytes((PCHAR) dataPointer, dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "AppendBytes():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   ::CoTaskMemFree(schema);
   ::CoTaskMemFree(dataPointer);

   return CString(printerStatusString);
}

////////////////////////////////////////////////////////////////////////////////
// util::GetPrinterStatusXML()
//
// Most commands are followed with a call to this function. It returns xml
// markup that contains the most recent state of the printer.
//
// This version of GetPrinterStatusXML() issues a dxp01sdk::PRINTER_MESSAGES
// request to get printer status xml.
////////////////////////////////////////////////////////////////////////////////
CString util::GetPrinterStatusXML(CComPtr <IBidiSpl>& bidiSpl) 
{
   stringstream utilExceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_MESSAGES);
   if (FAILED(hr)) {
      utilExceptionText << "SetSchema(dxp01sdk::PRINTER_MESSAGES): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_GET, bidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "SendRecv(BIDI_ACTION_GET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   HRESULT printerStatusHresult(S_OK);
   hr = bidiRequest->GetResult(&printerStatusHresult);
   if (FAILED(hr)) {
      utilExceptionText << "GetResult():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   LPWSTR schema(NULL);
   PBYTE dataPointer(NULL);
   DWORD dataType = 0;
   ULONG dataSizeInBytes = 0;
   hr = bidiRequest->GetOutputData(
      0,
      &schema,
      &dataType,
      &dataPointer,
      &dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "GetOutputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR printerMessagesXml;
   hr = printerMessagesXml.AppendBytes((PCHAR) dataPointer, dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "AppendBytes():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   ::CoTaskMemFree(schema);
   ::CoTaskMemFree(dataPointer);

   return CString(printerMessagesXml);
}

////////////////////////////////////////////////////////////////////////////////
// util::ParsePrinterStatusXML()
//
// for the given Printer Status XML string, populate and return a 
// PrinterStatusValues data structure.
//
// throw a std::runtime_error if any errors encountered.
////////////////////////////////////////////////////////////////////////////////
util::PrinterStatusValues util::ParsePrinterStatusXML(const CString printerStatusXML)
{
   // example Printer Status XML from printer:
   //
   // <?xml version="1.0" ?> 
   // <!-- Printer status xml file. --> 
   // <PrinterStatus>
   //   <ClientID>fellmadWin7x64_{545CF2E2-D68A-44B8-B83F-0DB07F21124F}</ClientID> 
   //   <WindowsJobID>0</WindowsJobID> 
   //   <PrinterJobID>8842</PrinterJobID> 
   //   <ErrorCode>0</ErrorCode> 
   //   <ErrorSeverity>0</ErrorSeverity> 
   //   <ErrorString /> 
   //   <DataFromPrinter>
   //     <![CDATA[]]>
   //   </DataFromPrinter>
   // </PrinterStatus>
   //
   // we populate a PrinterStatusValues struct with the values from this XML.
   // we use xpath to retrieve things.

   stringstream utilExceptionText;
   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(printerStatusXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty(SelectionLanguage):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   PrinterStatusValues printerStatusValues;

   printerStatusValues._clientID       = StringFromXml(doc, L"PrinterStatus/ClientID");
   printerStatusValues._errorCode      = IntFromXml(doc,    L"PrinterStatus/ErrorCode");
   printerStatusValues._errorSeverity  = IntFromXml(doc,    L"PrinterStatus/ErrorSeverity");
   printerStatusValues._errorString    = StringFromXml(doc, L"PrinterStatus/ErrorString");
   printerStatusValues._printerData    = StringFromXml(doc, L"PrinterStatus/DataFromPrinter");
   printerStatusValues._printerJobID   = IntFromXml(doc,    L"PrinterStatus/PrinterJobID");
   printerStatusValues._windowsJobID   = IntFromXml(doc,    L"PrinterStatus/WindowsJobID");

   return printerStatusValues;
}

CString util::FormatPrinterActionXML(const int actionID, const int printerJobID, const int errorCode)
{
   CString formattedPrinterActionXML;
   formattedPrinterActionXML.Format(dxp01sdk::PRINTER_ACTION_XML, actionID, printerJobID, errorCode);
   return formattedPrinterActionXML;
}

////////////////////////////////////////////////////////////////////////////////
// util::ParseMagstripeStrings()
//
// we're given a Printer Status XML fragment that has a CDATA section like this:
//
// example Printer Status XML from printer in the version 3.0 driver:
//
//<?xml version="1.0"?>
//<!--Printer status xml file.-->
//<PrinterStatus>
//   <ClientID>fellmadwin7x64_{8CB9F232-7136-42BC-B0FA-EF7D734EAEAD}</ClientID>
//   <WindowsJobID>0</WindowsJobID>
//   <PrinterJobID>6028</PrinterJobID>
//   <ErrorCode>0</ErrorCode>
//   <ErrorSeverity>0</ErrorSeverity>
//   <ErrorString></ErrorString>
//   <DataFromPrinter>
//       <![CDATA[<?xml version="1.0" encoding="UTF-8"?>
//       <DPCLMagStripe:MagStripe
//           ...
//           <track number="1">
//              <base64Data>VFJBQ0sx</base64Data>
//           </track>
//           <track number="2">
//              <base64Data>MTEyMg==</base64Data>
//           </track>
//           <track number="3">
//              <base64Data>MzIx</base64Data>
//           </track>
//       </DPCLMagStripe:MagStripe>]]>
//    </DataFromPrinter>
//</PrinterStatus>
//
// return three strings. the string data is still Base64 encoded.
////////////////////////////////////////////////////////////////////////////////
void util::ParseMagstripeStrings(
   const CString  printerStatusXML,
   CString&       track1,
   CString&       track2,
   CString&       track3,
   bool           bJISRequest)
{
   CComPtr <IXMLDOMDocument2> doc;
   stringstream utilExceptionText;

   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(printerStatusXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty(SelectionLanguage):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   // get the CDATA text() node:
   CComPtr <IXMLDOMNode> node;
   hr = doc->selectSingleNode(L"PrinterStatus/DataFromPrinter/text()", &node);
   if (FAILED(hr) || S_FALSE == hr) {
      utilExceptionText << "doc->selectSingleNode():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR cdataText;
   hr = node->get_text(&cdataText);
   if (FAILED(hr)) {
      utilExceptionText << "getting CDATA text:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   // create a new DOM doc from the CDATA text:
   doc = NULL;
   hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance() for CDATA:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->loadXML(cdataText, &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML() CDATA text:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComPtr <IXMLDOMNodeList> nodelist;
   hr = doc->selectNodes(L"magstripe/track", &nodelist);
   if (FAILED(hr)) {
      utilExceptionText << "doc->selectNodes(magstripe/track):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   if (nodelist) {
      long trackCount(0);
      hr = nodelist->get_length(&trackCount);
      if (FAILED(hr)) {
         utilExceptionText << "nodelist->get_length():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(utilExceptionText.str());
      }

      track1.Empty();
      track2.Empty();
      track3.Empty();

      for (int trackIndex = 0; trackIndex < trackCount; trackIndex++) {
         node = NULL;
         hr = nodelist->get_item(trackIndex, &node);
         if (FAILED(hr)) {
            utilExceptionText << "nodelist->get_item():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(utilExceptionText.str());
         }

         CComPtr <IXMLDOMNamedNodeMap> attributeMap;
         hr = node->get_attributes(&attributeMap);
         if (FAILED(hr)) {
            utilExceptionText << "node->get_attributes():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(utilExceptionText.str());
         }

         CComPtr <IXMLDOMNode> attributeNode;
         hr = attributeMap->getNamedItem(L"number", &attributeNode);
         if (FAILED(hr)) {
            utilExceptionText << "attributeMap->getNamedItem():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(utilExceptionText.str());
         }

         CComBSTR attributeText;
         hr = attributeNode->get_text(&attributeText);
         if (FAILED(hr)) {
            utilExceptionText << "attributeNode->get_text():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(utilExceptionText.str());
         }

         CComBSTR trackText;
         hr = node->get_text(&trackText);
         if (FAILED(hr)) {
            utilExceptionText << "node->get_text():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
            throw runtime_error(utilExceptionText.str());
         }

         if (!bJISRequest) {
            if (attributeText == L"1") {track1 = trackText; continue;}
            if (attributeText == L"2") {track2 = trackText; continue;}
         }

         if (attributeText == L"3") {track3 = trackText; continue;}
      }
   }
}

////////////////////////////////////////////////////////////////////////////////
// util::GetPrinterOptionsXML()
//
// get the 'printer options' xml from the printer. This is safe to call at any
// time; no print job is needed.
//
// Use ParsePrinterOptionsXML() to obtain a data structure with the values.
////////////////////////////////////////////////////////////////////////////////
CString util::GetPrinterOptionsXML(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream utilExceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_OPTIONS2);
   if (FAILED(hr)) {
      utilExceptionText << "SetSchema(dxp01sdk::PRINTER_OPTIONS2):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_GET, bidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "SendRecv(BIDI_ACTION_GET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   HRESULT printerStatusHresult(S_OK);
   hr = bidiRequest->GetResult(&printerStatusHresult);
   if (FAILED(hr)) {
      utilExceptionText << "GetResult():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   LPWSTR schema(NULL);
   PBYTE dataPointer(NULL);
   DWORD dataType = 0;
   ULONG dataSizeInBytes = 0;
   hr = bidiRequest->GetOutputData(
      0,
      &schema,
      &dataType,
      &dataPointer,
      &dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "GetOutputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR printerOptionsXML;
   hr = printerOptionsXML.AppendBytes((PCHAR) dataPointer, dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "AppendBytes:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   ::CoTaskMemFree(schema);
   ::CoTaskMemFree(dataPointer);

   return CString(printerOptionsXML);
}

////////////////////////////////////////////////////////////////////////////////
// util::GetPrinterCardCountsXML()
//
// get the 'printer card counts' xml from the printer. This is safe to call at any
// time; no print job is needed.
//
// this function just returns the XML in a string. Use ParsePrinterCardCountsXML()
// to obtain a data structure with the values.
////////////////////////////////////////////////////////////////////////////////
CString util::GetPrinterCardCountsXML(CComPtr <IBidiSpl>& bidiSpl)
{
   return GetPrinterActionReturnXML(bidiSpl, dxp01sdk::COUNTER_STATUS2);
}

CString util::GetSDKVersionXML(CComPtr <IBidiSpl>& bidiSpl)
{
   return GetPrinterActionReturnXML(bidiSpl, dxp01sdk::SDK_VERSION);
}

CString util::GetPrinterActionReturnXML(
   CComPtr <IBidiSpl>& bidiSpl,
   CString             actionString)
{
   stringstream utilExceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiRequest->SetSchema(actionString);
   if (FAILED(hr)) {
      utilExceptionText << "SetSchema(" << CW2A(actionString) << "): 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_GET, bidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "SendRecv(BIDI_ACTION_GET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   HRESULT printerStatusHresult(S_OK);
   hr = bidiRequest->GetResult(&printerStatusHresult);
   if (FAILED(hr)) {
      utilExceptionText << "GetResult():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   LPWSTR schema(NULL);
   PBYTE dataPointer(NULL);
   DWORD dataType = 0;
   ULONG dataSizeInBytes = 0;
   hr = bidiRequest->GetOutputData(
      0,
      &schema,
      &dataType,
      &dataPointer,
      &dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "GetOutputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR printerActionReturnXML;
   hr = printerActionReturnXML.AppendBytes((PCHAR) dataPointer, dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "AppendBytes:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   ::CoTaskMemFree(schema);
   ::CoTaskMemFree(dataPointer);

   return CString(printerActionReturnXML);
}

void util::WaitUntilJobSpooled(const HANDLE printerHandle, const DWORD windowsJobID)
{
   if (NULL == printerHandle)
      return;

   while (true)
   {
      // Wait for print data to make it to driver
      ::Sleep(1000);

      DWORD neededByteCount = 0;
      if (!::GetJob(printerHandle, windowsJobID, 2, NULL, 0, &neededByteCount))
      {
         if (::GetLastError() == ERROR_INSUFFICIENT_BUFFER)
         {
            vector <byte> jobInfo2Bytes(neededByteCount, 0);
            JOB_INFO_2* pJobInfo2 = (JOB_INFO_2*) (&jobInfo2Bytes[0]);

            DWORD returnedByteCount = 0;
            if (pJobInfo2 && ::GetJob(printerHandle, windowsJobID, 2, (byte*) pJobInfo2, neededByteCount, &returnedByteCount))
            {
               if (pJobInfo2->Status == JOB_STATUS_COMPLETE)
               {
                  break;                  
               }
            }
         }
      }
   }
}

void util::PrintTextAndGraphics(
   CString        printerName,
   const HANDLE   printerHandle,
   DWORD&         windowsJobID,
   const bool     duplex)
{
   stringstream utilExceptionText;

   const LONG devmode_size = ::DocumentProperties(
      NULL,
      printerHandle,
      printerName.GetBuffer(),
      NULL,
      NULL,
      0); // Set to 0 to return required buffer size.

   // this buffer will nicely self-destruct:
   vector <byte> devmodeBytes(devmode_size, 0);
   DEVMODE* devmode = (DEVMODE*) (&devmodeBytes[0]);

   LONG long_rc = ::DocumentProperties(
      NULL,
      printerHandle,
      printerName.GetBuffer(),
      devmode,
      NULL,
      DM_OUT_BUFFER);
   if (long_rc < 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "DocumentProperties():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   devmode->dmOrientation = DMORIENT_LANDSCAPE;

   if (duplex) {
      devmode->dmDuplex = DMDUP_VERTICAL;
   }
   else {
      devmode->dmDuplex = DMDUP_SIMPLEX;
   }

   HDC printerDC = ::CreateDC(
      L"WINSPOOL",
      devmode->dmDeviceName,
      NULL,
      devmode);
   if (NULL == printerDC) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "CreateDC():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   DOCINFO docInfo = {0};
   docInfo.cbSize = sizeof DOCINFO;
   docInfo.lpszDocName = L"xps printer driver magstripe document";
   int rc = ::StartDoc(printerDC, &docInfo);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "StartDoc():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   windowsJobID = rc;

   rc = ::StartPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "StartPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   CString text;
   text.Format(L"Sample printing on %s", printerName);
   rc = ::TextOut(printerDC, 50, 200, text, text.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   rc = ::EndPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "EndPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   if (duplex)
   {
      rc = ::StartPage(printerDC);
      if (rc <= 0) {
         const DWORD lastErr = ::GetLastError();
         utilExceptionText << "StartPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(utilExceptionText.str());
      }

      text.Format(L"Sample back side printing on %s", printerName);
      rc = ::TextOut(printerDC, 50, 200, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         utilExceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(utilExceptionText.str());
      }

      rc = ::EndPage(printerDC);
      if (rc <= 0) {
         const DWORD lastErr = ::GetLastError();
         utilExceptionText << "EndPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(utilExceptionText.str());
      }
   }

   rc = ::EndDoc(printerDC);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "EndDoc():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   rc = ::DeleteDC(printerDC);
   if (rc == 0) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "DeleteDC():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }
}

void util::PrintText(CString printerName, const bool duplex)
{
   stringstream utilExceptionText;
   HANDLE printerHandle(0);
   BOOL brc = ::OpenPrinter(printerName.GetBuffer(), &printerHandle, NULL);
   if (!brc) {
      const DWORD lastErr = ::GetLastError();
      utilExceptionText << "OpenPrinter():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(utilExceptionText.str());
   }

   // do some windows printing and get the windowsJobID:
   DWORD windowsJobID(0);
   PrintTextAndGraphics(printerName, printerHandle, windowsJobID, duplex);

   // Important: wait until driver gets all the print data:
   WaitUntilJobSpooled(printerHandle, windowsJobID); 
   ::ClosePrinter(printerHandle);
}

CString util::GetJobStatusXML(CComPtr <IBidiSpl>& bidiSpl, const DWORD printerJobID)
{
   stringstream utilExceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CString jobStatusRequestXML;
   jobStatusRequestXML.Format(dxp01sdk::JOB_STATUS_XML, printerJobID);

   ULONG dataLength = jobStatusRequestXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE jobStatusXMLBytes = (PBYTE) XMLBytesAllocator.Allocate(dataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(jobStatusXMLBytes, dataLength, jobStatusRequestXML.GetBuffer(), dataLength);

   // Setup the IBidiSpl request object using IBidiRequest ...
   hr = bidiRequest->SetSchema(dxp01sdk::JOB_STATUS);
   if (FAILED(hr)) {
      utilExceptionText << "SetSchema(dxp01sdk::JOB_STATUS):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   // send the request (xml that includes the printerJobID):
   hr = bidiRequest->SetInputData(BIDI_BLOB, jobStatusXMLBytes, dataLength);
   if (FAILED(hr)) {
      utilExceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   // Send the command to the printer.
   hr = bidiSpl->SendRecv(BIDI_ACTION_GET, bidiRequest);
   if (FAILED(hr)) {
      utilExceptionText << "bidiSpl->SendRecv(BIDI_ACTION_GET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   HRESULT printerStatusHresult(S_OK);
   hr = bidiRequest->GetResult(&printerStatusHresult);
   if (FAILED(hr)) {
      utilExceptionText << "GetResult():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   LPWSTR schema(NULL);
   PBYTE dataPointer(NULL);
   DWORD dataType = 0;
   ULONG dataSizeInBytes = 0;
   hr = bidiRequest->GetOutputData(
      0,
      &schema,
      &dataType,
      &dataPointer,
      &dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "GetOutputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CComBSTR jobStatusXML;
   hr = jobStatusXML.AppendBytes((PCHAR) dataPointer, dataSizeInBytes);
   if (FAILED(hr)) {
      utilExceptionText << "AppendBytes():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   ::CoTaskMemFree(schema);
   ::CoTaskMemFree(dataPointer);

   return CString(jobStatusXML);
}

////////////////////////////////////////////////////////////////////////////////
// util::ParseJobStatusXML()
//
// retrieve values from Job Status XML and populate a JobStatusValues data
// structure. Return the data structure.
//
// Job Status XML looks like this:
//
// <?xml version="1.0" ?> 
// <!-- Job status xml file. --> 
// <JobStatus>
//    <ClientID>fellmadwin7x64_{901F168C-B593-42B8-8602-4C715D8D658C}</ClientID> 
//    <WindowsJobID>0</WindowsJobID> 
//    <PrinterJobID>1327</PrinterJobID> 
//    <JobState>JobActive</JobState> 
//    <JobRestartCount>0</JobRestartCount> 
// </JobStatus>
//
////////////////////////////////////////////////////////////////////////////////
util::JobStatusValues util::ParseJobStatusXML(const CString jobStatusXML)
{
   stringstream utilExceptionText;

   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(jobStatusXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   JobStatusValues jobStatusValues;

   jobStatusValues._clientID        = StringFromXml(doc, L"JobStatus/ClientID");
   jobStatusValues._jobState        = StringFromXml(doc, L"JobStatus/JobState");
   jobStatusValues._windowsJobID    = IntFromXml(doc,    L"JobStatus/WindowsJobID");
   jobStatusValues._printerJobID    = IntFromXml(doc,    L"JobStatus/PrinterJobID");
   jobStatusValues._jobRestartCount = IntFromXml(doc,    L"JobStatus/JobRestartCount");

   return jobStatusValues;
}

////////////////////////////////////////////////////////////////////////////////
// util::ParsePrinterOptionsXML()
//
// parse the xml returned from GetPrinterOptionsXML(). Populate a
// PrinterOptionsValues data structure with the values and return it.
//
// sample xml markup returned from printer sorted by element name:
//
//    <?xml version="1.0"?>
//    <!--Printer options2 xml file.-->
//    <PrinterInfo2>
//       <ColorPrintResolution>300x300 | 300x600</ColorPrintResolution>
//       <ConnectionPortType>Network</ConnectionPortType>
//       <ConnectionProtocol>Version2Secure</ConnectionProtocol>
//       <EmbosserVersion>E1.1.25-0</EmbosserVersion>
//       <LockState>Locked</LockState>
//       <ModuleEmbosser>Installed</ModuleEmbosser>
//       <MonochromePrintResolution>300x300 | 300x600 | 300x1200</MonochromePrintResolution>
//       <OptionDuplex>Auto</OptionDuplex>
//       <OptionInputhopper>MultiHopper6</OptionInputhopper>
//       <OptionLocks>Installed</OptionLocks>
//       <OptionMagstripe>ISO</OptionMagstripe>
//       <OptionSmartcard>Single wire</OptionSmartcard>
//       <PrinterAddress>172.16.2.211</PrinterAddress>
//       <PrinterMessageNumber>0</PrinterMessageNumber>
//       <PrinterModel>CD870</PrinterModel>
//       <PrinterSerialNumber>E00105</PrinterSerialNumber>
//       <PrinterStatus>Ready</PrinterStatus>
//       <PrinterVersion>D3.12.4-0</PrinterVersion>
//       <PrintHead>Installed</PrintHead>
//       <TopcoatPrintResolution>300x300</TopcoatPrintResolution>
//    </PrinterInfo2>
//
////////////////////////////////////////////////////////////////////////////////
util::PrinterOptionsValues util::ParsePrinterOptionsXML(const CString printerOptionsXML)
{
   stringstream utilExceptionText;

   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(printerOptionsXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   PrinterOptionsValues printerOptionsValues;

   printerOptionsValues._colorPrintResolution      = StringFromXml(doc, L"PrinterInfo2/ColorPrintResolution");
   printerOptionsValues._connectionPortType        = StringFromXml(doc, L"PrinterInfo2/ConnectionPortType");
   printerOptionsValues._connectionProtocol        = StringFromXml(doc, L"PrinterInfo2/ConnectionProtocol");
   printerOptionsValues._embosserVersion           = StringFromXml(doc, L"PrinterInfo2/EmbosserVersion");
   printerOptionsValues._lockState                 = StringFromXml(doc, L"PrinterInfo2/LockState");
   printerOptionsValues._moduleEmbosser            = StringFromXml(doc, L"PrinterInfo2/ModuleEmbosser");
   printerOptionsValues._monochromePrintResolution = StringFromXml(doc, L"PrinterInfo2/MonochromePrintResolution");
   printerOptionsValues._optionDuplex              = StringFromXml(doc, L"PrinterInfo2/OptionDuplex");
   printerOptionsValues._optionInputhopper         = StringFromXml(doc, L"PrinterInfo2/OptionInputhopper");
   printerOptionsValues._optionLocks               = StringFromXml(doc, L"PrinterInfo2/OptionLocks");
   printerOptionsValues._optionMagstripe           = StringFromXml(doc, L"PrinterInfo2/OptionMagstripe");
   printerOptionsValues._optionSmartcard           = StringFromXml(doc, L"PrinterInfo2/OptionSmartcard");
   printerOptionsValues._printerAddress            = StringFromXml(doc, L"PrinterInfo2/PrinterAddress");
   printerOptionsValues._printerMessageNumber      = IntFromXml(doc,    L"PrinterInfo2/PrinterMessageNumber");
   printerOptionsValues._printerModel              = StringFromXml(doc, L"PrinterInfo2/PrinterModel");
   printerOptionsValues._printerSerialNumber       = StringFromXml(doc, L"PrinterInfo2/PrinterSerialNumber");
   printerOptionsValues._printerStatus             = StringFromXml(doc, L"PrinterInfo2/PrinterStatus");
   printerOptionsValues._printerVersion            = StringFromXml(doc, L"PrinterInfo2/PrinterVersion");
   printerOptionsValues._printHead                 = StringFromXml(doc, L"PrinterInfo2/PrintHead");
   printerOptionsValues._topcoatPrintResolution    = StringFromXml(doc, L"PrinterInfo2/TopcoatPrintResolution");

   return printerOptionsValues;
}

////////////////////////////////////////////////////////////////////////////////
// util::ParsePrinterCardCountsXML()
//
// the 'printer card counts' xml from the printer looks something like this:
//
//    <?xml version="1.0"?>
//    <!--Printer counter2 xml file.-->
//    <CounterStatus2>
//       <CardsPickedSinceCleaningCard>141</CardsPickedSinceCleaningCard>
//       <CleaningCardsRun>0</CleaningCardsRun>
//       <CurrentCompleted>260</CurrentCompleted>
//       <CurrentLost>0</CurrentLost>
//       <CurrentPicked>474</CurrentPicked>
//       <CurrentPickedExceptionSlot>2</CurrentPickedExceptionSlot>
//       <CurrentPickedInputHopper1>242</CurrentPickedInputHopper1>
//       <CurrentPickedInputHopper2>1</CurrentPickedInputHopper2>
//       <CurrentPickedInputHopper3>33</CurrentPickedInputHopper3>
//       <CurrentPickedInputHopper4>14</CurrentPickedInputHopper4>
//       <CurrentPickedInputHopper5>3</CurrentPickedInputHopper5>
//       <CurrentPickedInputHopper6>1</CurrentPickedInputHopper6>
//       <CurrentRejected>36</CurrentRejected>
//       <PrinterStatus>Ready</PrinterStatus>
//       <TotalCompleted>260</TotalCompleted>
//       <TotalLost>0</TotalLost>
//       <TotalPicked>474</TotalPicked>
//       <TotalPickedExceptionSlot>2</TotalPickedExceptionSlot>
//       <TotalPickedInputHopper1>242</TotalPickedInputHopper1>
//       <TotalPickedInputHopper2>1</TotalPickedInputHopper2>
//       <TotalPickedInputHopper3>33</TotalPickedInputHopper3>
//       <TotalPickedInputHopper4>14</TotalPickedInputHopper4>
//       <TotalPickedInputHopper5>3</TotalPickedInputHopper5>
//       <TotalPickedInputHopper6>1</TotalPickedInputHopper6>
//       <TotalRejected>36</TotalRejected>
//    </CounterStatus2>
//
// parse the xml returned from GetPrinterCardCountsXML(). Populate a
// PrinterCardCountValues data structure with the values and return it.
//
////////////////////////////////////////////////////////////////////////////////
util::PrinterCardCountValues util::ParsePrinterCardCountsXML(const CString printerCardCountsXML)
{
   stringstream utilExceptionText;
   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(printerCardCountsXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty(SelectionLanguage):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   PrinterCardCountValues printerCardCountValues;

   printerCardCountValues._cardsPickedSinceCleaningCard  = IntFromXml(doc,    L"CounterStatus2/CardsPickedSinceCleaningCard");
   printerCardCountValues._cleaningCardsRun              = IntFromXml(doc,    L"CounterStatus2/CleaningCardsRun");
   printerCardCountValues._currentCompleted              = IntFromXml(doc,    L"CounterStatus2/CurrentCompleted");
   printerCardCountValues._currentLost                   = IntFromXml(doc,    L"CounterStatus2/CurrentLost");
   printerCardCountValues._currentPicked                 = IntFromXml(doc,    L"CounterStatus2/CurrentPicked");
   printerCardCountValues._currentPickedExceptionSlot    = IntFromXml(doc,    L"CounterStatus2/CurrentPickedExceptionSlot");
   printerCardCountValues._currentPickedInputHopper1     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper1");
   printerCardCountValues._currentPickedInputHopper2     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper2");
   printerCardCountValues._currentPickedInputHopper3     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper3");
   printerCardCountValues._currentPickedInputHopper4     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper4");
   printerCardCountValues._currentPickedInputHopper5     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper5");
   printerCardCountValues._currentPickedInputHopper6     = IntFromXml(doc,    L"CounterStatus2/CurrentPickedInputHopper6");
   printerCardCountValues._currentRejected               = IntFromXml(doc,    L"CounterStatus2/CurrentRejected");
   printerCardCountValues._printerStatus                 = StringFromXml(doc, L"CounterStatus2/PrinterStatus");
   printerCardCountValues._totalCompleted                = IntFromXml(doc,    L"CounterStatus2/TotalCompleted");
   printerCardCountValues._totalLost                     = IntFromXml(doc,    L"CounterStatus2/TotalLost");
   printerCardCountValues._totalPicked                   = IntFromXml(doc,    L"CounterStatus2/TotalPicked");
   printerCardCountValues._totalPickedExceptionSlot      = IntFromXml(doc,    L"CounterStatus2/TotalPickedExceptionSlot");
   printerCardCountValues._totalPickedInputHopper1       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper1");
   printerCardCountValues._totalPickedInputHopper2       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper2");
   printerCardCountValues._totalPickedInputHopper3       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper3");
   printerCardCountValues._totalPickedInputHopper4       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper4");
   printerCardCountValues._totalPickedInputHopper5       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper5");
   printerCardCountValues._totalPickedInputHopper6       = IntFromXml(doc,    L"CounterStatus2/TotalPickedInputHopper6");
   printerCardCountValues._totalRejected                 = IntFromXml(doc,    L"CounterStatus2/TotalRejected");

   return printerCardCountValues;  
}

CString util::ParseSDKVersionXML(const CString sdkXML)
{
   // sample sdk version xml:
   //
   //    <?xml version="1.0"?>
   //    <!--Printer SDK version xml file.-->
   //    <SDKVersion>3.0.258.0</SDKVersion>
   //
   stringstream utilExceptionText;
   
   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(sdkXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   CString sdkVersionString = util::StringFromXml(doc, L"SDKVersion");
   return sdkVersionString;
}

////////////////////////////////////////////////////////////////////////////////
// util::ParseSmartcardResponseXML()
//
// incoming XML looks something like this...from the CDATA section of a
// PrinterStatus response:
//
//    <?xml version="1.0" ?> 
//    <!-- smartcard response xml  --> 
//    <SmartcardResponse>
//       <Protocol>SCARD_PROTOCOL_RAW</Protocol> 
//       <State>SCARD_PRESENT | SCARD_POWERED | SCARD_NEGOTIABLE</State> 
//       <Status>SCARD_S_SUCCESS</Status> 
//       <Base64Data>O4WBAQEay1xPxg==</Base64Data> 
//    </SmartcardResponse>
//
// return a SmartcardResponseValues populated with those items.
////////////////////////////////////////////////////////////////////////////////
util::SmartcardResponseValues util::ParseSmartcardResponseXML(
   const CString smartcardReponseXML)
{
   stringstream utilExceptionText;
   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(smartcardReponseXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   util::SmartcardResponseValues smartcardResponseValues;
   smartcardResponseValues._protocol = util::StringFromXml(doc, L"SmartcardResponse/Protocol").Trim();

   CString statesString = util::StringFromXml(doc, L"SmartcardResponse/State").Trim();
   CString token;
   int currentPosition(0);
   do {
      token = statesString.Tokenize(L" |", currentPosition);
      if (!token.IsEmpty()) {
         wstring stateString = token;
         smartcardResponseValues._states.push_back(stateString);
      }
   } while (!token.IsEmpty());

   smartcardResponseValues._status = util::StringFromXml(doc, L"SmartcardResponse/Status").Trim();
   smartcardResponseValues._base64string = util::StringFromXml(doc, L"SmartcardResponse/Base64Data").Trim();

   if (smartcardResponseValues._base64string.length() > 0)
   {
      int convertedLength = (int) smartcardResponseValues._base64string.length();
      smartcardResponseValues._bytesFromBase64String.resize(convertedLength);
      ATL::Base64Decode(
         CW2A(smartcardResponseValues._base64string.c_str()),
         (int) smartcardResponseValues._base64string.length(),
         &smartcardResponseValues._bytesFromBase64String[0],
         &convertedLength);
      smartcardResponseValues._bytesFromBase64String.resize(convertedLength);
   }
   return smartcardResponseValues;
}

CString util::GetSuppliesXML(CComPtr <IBidiSpl>& bidiSpl)
{
   CString suppliesStatusXML = GetPrinterActionReturnXML(bidiSpl, dxp01sdk::SUPPLIES_STATUS2);
   return suppliesStatusXML;
}

util::SuppliesValues util::ParseSuppliesXML(const CString suppliesXML)
{
   // sample supplies xml from printer:
   //
   //    <?xml version="1.0"?>
   //    <!--Printer Supplies xml file.-->
   //    <PrinterSupplies>
   //       <PrinterStatus>Ready</PrinterStatus>
   //       <PrintRibbonType>YMCKTKT</PrintRibbonType>
   //       <RibbonRemaining>100</RibbonRemaining>
   //       <IndentRibbon>Installed</IndentRibbon>
   //       <IndentRibbonRemaining>20</IndentRibbonRemaining>
   //       <TopperRibbonType>Blue</TopperRibbonType>
   //       <TopperRibbonRemaining>94</TopperRibbonRemaining>
   //    </PrinterSupplies>

   stringstream utilExceptionText;

   CComPtr <IXMLDOMDocument2> doc;
   HRESULT hr = doc.CoCreateInstance(__uuidof(DOMDocument60)); 
   if (FAILED(hr)) {
      utilExceptionText << "doc.CoCreateInstance():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->put_async(false);
   if (FAILED(hr)) {
      utilExceptionText << "doc->put_async():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   VARIANT_BOOL loadedOK(VARIANT_FALSE);
   hr = doc->loadXML(CComBSTR(suppliesXML), &loadedOK);
   if (FAILED(hr)) {
      utilExceptionText << "doc->loadXML():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   hr = doc->setProperty(L"SelectionLanguage", CComVariant(L"XPath"));
   if (FAILED(hr)) {
      utilExceptionText << "doc->setProperty():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(utilExceptionText.str());
   }

   SuppliesValues suppliesValues;

   suppliesValues._indentRibbon           = StringFromXml(doc,       L"PrinterSupplies2/IndentRibbon");
   suppliesValues._indentRibbonRemaining  = (short) IntFromXml(doc,  L"PrinterSupplies2/IndentRibbonRemaining");
   suppliesValues._printerStatus          = StringFromXml(doc,       L"PrinterSupplies2/PrinterStatus");
   suppliesValues._printRibbonType        = StringFromXml(doc,       L"PrinterSupplies2/PrintRibbonType");
   suppliesValues._ribbonRemaining        = (short) IntFromXml(doc,  L"PrinterSupplies2/RibbonRemaining");
   suppliesValues._printRibbonSerialNumber= StringFromXml(doc,       L"PrinterSupplies2/RibbonSerialNumber");
   suppliesValues._printRibbonPartNumber  = (long) IntFromXml(doc,   L"PrinterSupplies2/RibbonPartNumber");
   suppliesValues._printRibbonLotCode     = StringFromXml(doc,       L"PrinterSupplies2/RibbonLotCode");
   suppliesValues._topperRibbonRemaining  = (short) IntFromXml(doc,  L"PrinterSupplies2/TopperRibbonRemaining");
   suppliesValues._topperRibbonType       = StringFromXml(doc,       L"PrinterSupplies2/TopperRibbonType");
   
   return suppliesValues;
}
