////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// XPS Printer Driver SDK: emboss + indent + hopper selection c++ sample.
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <bidispl.h>
#include <WinSpool.h>
#include <iostream>
#include <sstream>
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

void usage(CString longExeName)
{
   CPath shortExeName(longExeName);
   shortExeName.StripPath();
   cout << endl;
   cout << CT2A(shortExeName) << " demonstrates the emboss and indent escapes, input hopper" << endl;
   cout << "selection, and the checking of print ribbon and emboss supplies." << endl << endl;
   cout << "Emboss and indent data is hard coded." << endl << endl;
   cout << CT2A(shortExeName) << " -n <printername> [-i <input hopper>] [-r] [-e] [-c]" << endl << endl;
   cout << "options:" << endl;
   cout << "  " << "-n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  " << "-i <input hopper>. Optional. Defaults to \"1\"." << endl;
   cout << "  " << "-r Check Print Ribbon during StartJob. Optional. Defaults to false." << endl;
   cout << "  " << "-e Check Emboss Supplies during StartJob. Optional. Defaults to false." << endl;
   cout << "  " << "-c Poll for job status. Optional. Defaults to false." << endl;
   exit(0);
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

void CancelCard(CComPtr <IBidiSpl>& bidiSpl, const util::PrinterStatusValues& printerStatusValues)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

////////////////////////////////////////////////////////////////////////////////
// StartJob()
// issue an IBidi startjob.
////////////////////////////////////////////////////////////////////////////////
int StartJob(
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           checkPrintRibbon,
   const bool           cheeckEmbossSupplies,
   const wstring        inputHopperName = L"1")
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString startjobXml;
   startjobXml.Format(dxp01sdk::STARTJOB_XML, inputHopperName.c_str(), checkPrintRibbon, cheeckEmbossSupplies);
   const ULONG startjobDataLength = startjobXml.GetLength() * sizeof(TCHAR);

   // Get memory that can be used with IBidiSpl:
   CComAllocator XMLBytesAllocator; // automatically frees COM memory
   PBYTE XMLBytes = (PBYTE) XMLBytesAllocator.Allocate(startjobDataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(XMLBytes, startjobDataLength, startjobXml.GetBuffer(), startjobDataLength);

   // pass the startjob input data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, XMLBytes, startjobDataLength);
   if (FAILED(hr)) {
      exceptionText << "SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      // the StartJob failed; send a Cancel(); do NOT send an EndJob():
      CancelCard(bidiSpl, printerStatusValues);
      exceptionText << "startjob(): " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }

   return printerStatusValues._printerJobID;
}

////////////////////////////////////////////////////////////////////////////////
// PollForJobCompletion()
//
// repeatedly check job status until the job is complete.
////////////////////////////////////////////////////////////////////////////////
void PollForJobCompletion(
   CComPtr <IBidiSpl>&  bidiSpl,
   const int            printerJobID,
   const int            intervalMilliseconds)
{
   for (; true ;) {
      const CString jobStatusXML = util::GetJobStatusXML(bidiSpl, printerJobID);
      const util::JobStatusValues jobStatusValues = util::ParseJobStatusXML(jobStatusXML);
      cout << "Printer job id " << printerJobID << " " << CW2A(jobStatusValues._jobState) << endl;
      if (dxp01sdk::JOB_ACTIVE != jobStatusValues._jobState) {
         return;
      }
      ::Sleep(intervalMilliseconds);
   }
}

////////////////////////////////////////////////////////////////////////////////
// SendEmbossIndentData()
// use escapes in a windows print job document to emboss and indent a card.
////////////////////////////////////////////////////////////////////////////////
void SendEmbossIndentData(CString printerName)
{
   stringstream exceptionText;

   HANDLE printerHandle(0);  // needed for job completion
   BOOL brc = ::OpenPrinter(printerName.GetBuffer(), &printerHandle, NULL);
   if (!brc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "OpenPrinter():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   ::HDC printerDC = ::CreateDC(L"WINSPOOL", printerName, NULL, NULL);
   if (NULL == printerDC) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "CreateDC():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   ::DOCINFO docInfo = {0};
   docInfo.cbSize = sizeof DOCINFO;
   docInfo.lpszDocName = L"emboss indent job";
   int rc = ::StartDoc(printerDC, &docInfo);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "StartDoc():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   const DWORD windowsJobID = rc; // needed for job completion

   rc = ::StartPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "StartPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CString text;

   // emit all emboss + indent data driver escape strings:
   {
      text = L"~EM%1;301;860;Font 11111";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%2;1600;860;222222";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%3;301;1460;333333";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%4;301;1180;444444";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%5;301;690;555555";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%6;1600;690;666666";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%7;301;650;777777";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%8;301;1000;888888";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%9;301;1050;999999";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
      text = L"~EM%10;1600;1050;10 10 10";
      rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
      if (0 == rc) {
         const DWORD lastErr = ::GetLastError();
         exceptionText << "TextOut():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
         throw runtime_error(exceptionText.str());
      }
   }

   rc = ::EndPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "EndPage():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = ::EndDoc(printerDC);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "EndDoc():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = ::DeleteDC(printerDC);
   if (rc == 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "DeleteDC():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   util::WaitUntilJobSpooled(printerHandle, windowsJobID);

   ::ClosePrinter(printerHandle);
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   stringstream exceptionText;
   CString printerName;
   wstring inputHopper(L"1");
   bool checkPrintRibbon(false);
   bool checkEmbossSupplies(false);
   bool pollForJobCompletion(false);
   int c(0);

   while ((c = getopt(argc, argv, _T("n:ci:re"))) != EOF) {
      switch (c) {
      case L'n':
         printerName = optarg; break;
      case L'i':
         inputHopper = optarg; break;
      case L'c':
         pollForJobCompletion = true; break;
      case L'r':
         checkPrintRibbon = true; break;
      case L'e':
         checkEmbossSupplies = true; break;
      default:
         usage(argv[0]);
      }
   }

   if (printerName.IsEmpty()) {
      usage(argv[0]);
   }

   CComPtr <IBidiSpl> bidiSpl;
   try {
      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // create an IBidiSpl COM smartpointer instance and bind it to the printer
      // name. We need it for job completion status - not embossing or indenting.
      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(printerName) << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      const CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      const CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      // is the printer is in the ready state?
      const CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);
      if ("Ready" != printerOptionsValues._printerStatus) {
         exceptionText << "printer " << CT2A(printerName) << " is not ready.";
         throw runtime_error(exceptionText.str());
      }

      const int printerJobID = StartJob(
         bidiSpl,
         checkPrintRibbon,
         checkEmbossSupplies,
         inputHopper);

      SendEmbossIndentData(printerName);

      EndJob(bidiSpl);

      if (pollForJobCompletion) {
         // Embossing and indenting is slow. Check every 5 seconds for job
         // completion:
         PollForJobCompletion(bidiSpl, printerJobID, 5000);
      }
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}
