////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <vector>
#include <iostream>
#include <sstream>
#include <bidispl.h>
#include <atlbase.h>
#include <atlpath.h>
#include <atlstr.h>
#include <atlenc.h>
#include "util.h"
#include "XGetopt.h"
#include "DXP01SDK.H"

using namespace std;

void usage(CString longExeName)
{
   CPath shortExeName(longExeName);
   shortExeName.StripPath();

   cout << endl;
   cout << CT2A(shortExeName) << " demonstrates interactive mode magnetic stripe encoding with" << endl;
   cout << "options to read the magnetic stripe data, print, and poll for job completion" << endl;
   cout << "status." << endl << endl;
   cout << "Uses hard coded data for magnetic stripe and printing." << endl << endl;
   cout << CT2A(shortExeName) << " -n <printername> [-r] [-p] [-c] [-j] [-v]" << endl << endl;
   cout << "options:" << endl;
   cout << "  -n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  -r Reads back the encoded magnetic stripe data." << endl;
   cout << "  -p Print simple black text on one or both sides of the card"  << endl; 
   cout << "     depending on the printer used."  << endl; 
   cout << "  -c Poll for job status; needed to check for magstripe errors." << endl;
   cout << "  -j JIS magnetic." << endl;
   cout << "  -v verbose mode with timing information." << endl;

   shortExeName.RemoveExtension();

   cout << endl << CT2A(shortExeName) << " -n \"XPS Card Printer\"" << endl;
   cout << "Encodes data on all three tracks of an ISO 3-track magnetic stripe on" << endl;
   cout << "the backside of a card." << endl << endl;
   cout << CT2A(shortExeName) << " -n \"XPS Card Printer\" -r -p -c" << endl;
   cout << "Encodes data on all three tracks of an ISO 3-track magnetic stripe on" << endl;
   cout << "the backside of a card, reads the data from the magnetic stripe and" << endl;
   cout << "displays it, prints black text on one or both sides of the card," << endl;
   cout << "and polls and displays job status." << endl;
   cout << endl;
   exit(0);
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EndJob(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EndJob(): bidiSpl->SendRecv() returned");}
}

void CancelCard(
   CComPtr <IBidiSpl>& bidiSpl,
   const util::PrinterStatusValues& printerStatusValues)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   PBYTE xmlBytes = (PBYTE) ::CoTaskMemAlloc(xmlBytesLength);

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   ::CoTaskMemFree(xmlBytes);
}

int StartJob(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString startjobXml;
   startjobXml.Format(
      dxp01sdk::STARTJOB_XML,
      "",   // default input hopper
      0,    // zero means "do not check print ribbon"
      0);   // zero means "do not check emboss supplies"

   const ULONG startjobDataLength = startjobXml.GetLength() * sizeof(TCHAR);

   // Get memory that can be used with IBidiSpl:
   CComAllocator XMLBytesAllocator; // automatically frees COM memory
   PBYTE XMLBytes = (PBYTE) XMLBytesAllocator.Allocate(startjobDataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(XMLBytes, startjobDataLength, startjobXml.GetBuffer(), startjobDataLength);

   // pass the startjob input data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, XMLBytes, startjobDataLength);
   if (FAILED(hr)) {
      exceptionText << "SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("StartJob(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("StartJob(): bidiSpl->SendRecv() return");}

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues);
      exceptionText << "startjob():" << " 0x" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("returning from startjob()");}

   return printerStatusValues._printerJobID;
}

bool EncodeMagstripe(
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           JISRequest,
   const bool           verbose)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // hardcoded XML to encode all 3 tracks in IAT mode.
   // track 1 = "TRACK1", track 2 = "1122", track 3 = "321"
   CString trackDataXML = L"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
      L"<magstripe>"
      L"<track number=\"1\"><base64Data>VFJBQ0sx</base64Data></track>"
      L"<track number=\"2\"><base64Data>MTEyMg==</base64Data></track>"
      L"<track number=\"3\"><base64Data>MzIx</base64Data></track>"
      L"</magstripe>";

   if (JISRequest)
   {
      // JIS only allows track 3 = "321"
      trackDataXML = L"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
         L"<magstripe>"
         L"<track number=\"1\"><base64Data></base64Data></track>"
         L"<track number=\"2\"><base64Data></base64Data></track>"
         L"<track number=\"3\"><base64Data>MzIx</base64Data></track>"
         L"</magstripe>";
   }

   // Data is the length of the string multiplied by size of wide char
   ULONG dataLength = trackDataXML.GetLength() * sizeof WCHAR;

   // Get memory that can be used with IBidiSpl
   PBYTE trackDataXMLBytes = (PBYTE) ::CoTaskMemAlloc(dataLength);

   // copy the XML string...but not the terminating null:
   ::memcpy_s(trackDataXMLBytes, dataLength, trackDataXML.GetBuffer(), dataLength);

   // Setup the IBidiSpl request object using IBidiRequest ...
   // First, we are going to encode a magstripe.
   // replace schema string dxp01sdk::MAGSTRIPE_ENCODE to dxp01sdk::MAGSTRIPE_ENCODE_FRONT for front side encode
   hr = bidiRequest->SetSchema(dxp01sdk::MAGSTRIPE_ENCODE);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::MAGSTRIPE_ENCODE):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // pass the magstripe data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, trackDataXMLBytes, dataLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EncodeMagStripe(): SendRecv(BIDI_ACTION_SET) magstripe data as BIDI_BLOB");}

   // Send the command to the printer.
   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiSpl->SendRecv(BIDI_ACTION_SET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues);
      exceptionText << "EncodeMagStripe():" << " 0x" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   // Free the memory that was allocated for the IBidiSpl SendRecv call:
   ::CoTaskMemFree(trackDataXMLBytes);

   if (verbose) {util::EchoCurrentTime("EncodeMagStripe(): return");}

   return true;
}

////////////////////////////////////////////////////////////////////////////////
// PollForJobCompletion()
//
// repeatedly check job status until the job is complete.
////////////////////////////////////////////////////////////////////////////////
void PollForJobCompletion(
   CComPtr <IBidiSpl>&  bidiSpl,
   const int            printerJobID,
   const bool           verbose)
{
   while (true) {

      if (verbose) {util::EchoCurrentTime("PollForJobCompletion(): calling util::GetPrinterStatusXML()");}

      const CString printerStatusXml = util::GetPrinterStatusXML(bidiSpl);
      const util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXml);

      if (0 != printerStatusValues._errorCode) {
         cerr << CW2A(printerStatusValues._errorString) << " severity: " << printerStatusValues._errorSeverity << endl;
         cerr << "canceling printer jobID " << printerStatusValues._printerJobID << endl;
         CancelCard(bidiSpl, printerStatusValues);
         return;
      }

      const CString jobStatusXML = util::GetJobStatusXML(bidiSpl, printerJobID);
      const util::JobStatusValues jobStatusValues = util::ParseJobStatusXML(jobStatusXML);

      cout << "Printer job id " << printerJobID << " " << CW2A(jobStatusValues._jobState) << endl;

      if (dxp01sdk::JOB_ACTIVE != jobStatusValues._jobState) {
         return;
      }

      ::Sleep(1000); // milliseconds
   }
}

void ReadMagstripe(
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           bJISRequest, 
   const bool           verbose)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // replace schema string dxp01sdk::MAGSTRIPE_READ to dxp01sdk::MAGSTRIPE_READ_FRONT for front side read
   hr = bidiRequest->SetSchema(dxp01sdk::MAGSTRIPE_READ);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetSchema(dxp01sdk::MAGSTRIPE_READ):"
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ReadMagstripe(): calling bidiSpl->SendRecv(BIDI_ACTION_GET) to read magstripe data");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_GET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiSpl->SendRecv(BIDI_ACTION_GET):"
         << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ReadMagstripe(): bidiSpl->SendRecv(BIDI_ACTION_GET) returned");}

   // Check if device encoded data and SendRecv was a success
   HRESULT MagstripeEncodeResult(S_OK);
   hr = bidiRequest->GetResult(&MagstripeEncodeResult);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->GetResult():"<< " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(MagstripeEncodeResult)) {
      exceptionText << "MagstripeEncodeResult code:"<< " 0x" << hex << MagstripeEncodeResult
         << " " << CT2A(util::Win32ErrorString(MagstripeEncodeResult));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ReadMagstripe(): GetPrinterStatusXML()");}

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues);
      exceptionText << "ReadMagStripe():" << " 0x" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ReadMagstripe(): GetPrinterStatusXML() returned");}

   CString track1;
   CString track2;
   CString track3;

   util::ParseMagstripeStrings(printerStatusXML, track1, track2, track3, bJISRequest);

   int trackBytesLength = track1.GetLength();
   if (trackBytesLength) {
      wcout << L"track1:" << L" " << track1.GetBuffer() << endl;
      vector <byte> trackBytes(trackBytesLength);
      ::Base64Decode(CT2A(track1),  trackBytesLength, &trackBytes[0], &trackBytesLength);
      string s;
      s.append((char*) &trackBytes[0], trackBytesLength);
      cout << "track1 Base64 decoded: " << s << endl << endl;
   }
   else cout << "no track1 data read." << endl;

   trackBytesLength = track2.GetLength();
   if (trackBytesLength) {
      wcout << L"track2:" << L" " << track2.GetBuffer() << endl;
      vector <byte> trackBytes(trackBytesLength);
      ::Base64Decode(CT2A(track2),  trackBytesLength, &trackBytes[0], &trackBytesLength);
      string s;
      s.append((char*) &trackBytes[0], trackBytesLength);
      cout << "track2 Base64 decoded: " << s << endl << endl;
   }
   else cout << "no track2 data read." << endl;

   trackBytesLength = track3.GetLength();
   if (trackBytesLength) {
      wcout << L"track3:" << L" " << track3.GetBuffer() << endl;
      vector <byte> trackBytes(trackBytesLength);
      ::Base64Decode(CT2A(track3),  trackBytesLength, &trackBytes[0], &trackBytesLength);
      string s;
      s.append((char*) &trackBytes[0], trackBytesLength);
      cout << "track3 Base64 decoded: " << s << endl << endl;
   }
   else cout << "no track3 data read." << endl;
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   bool print(false);
   bool readMagstripe(false);
   bool pollForJobCompletion(false);
   bool jisRequest(false);
   bool verbose(false);
   CString printerName;

   int c(0);
   while ((c = getopt(argc, argv, _T("jpcrn:v"))) != EOF)
   {
      switch (c)
      {
      case TEXT('n'): printerName = optarg; break;
      case TEXT('r'): readMagstripe = true; break;
      case TEXT('p'): print = true; break;
      case TEXT('c'): pollForJobCompletion = true; break;
      case TEXT('j'): jisRequest = true; break;
      case TEXT('v'): verbose = true; break;
      default: usage(argv[0]);
      }
   }

   if (printerName.IsEmpty()) {
      usage(argv[0]);
   }

   if (verbose) {util::EchoCurrentTime("start");}

   CComPtr <IBidiSpl> bidiSpl;

   try
   {
      stringstream exceptionText;

      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(printerName) << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      // see if the printer is in the Printer_Ready state:
      if (verbose) {util::EchoCurrentTime("calling GetPrinterOptionsXML()");}
      CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);

      if ("Ready" != printerOptionsValues._printerStatus) {
         exceptionText << "printer " << CT2A(printerName) << " is not ready.";
         throw runtime_error(exceptionText.str());
      }

      int printerJobID = StartJob(bidiSpl, verbose);

      EncodeMagstripe(bidiSpl, jisRequest, verbose);

      if (readMagstripe) {
         ReadMagstripe(bidiSpl, jisRequest, verbose);
      }

      if (print) {
         // do some simple text and graphics printing.

         const bool doTwoSided = 
            printerOptionsValues._optionDuplex == "Auto" ||
            printerOptionsValues._optionDuplex == "Manual";

         if (verbose) {util::EchoCurrentTime("calling util::PrintText()");}

         // Important: util::PrintTextmethod also waits until driver gets all the print data.
         util::PrintText(printerName, doTwoSided);

         if (verbose) {util::EchoCurrentTime("util::PrintText() returned");}
      }

      EndJob(bidiSpl, verbose);

      if (pollForJobCompletion) {
         PollForJobCompletion(bidiSpl, printerJobID, verbose);
      }
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      if (verbose) {util::EchoCurrentTime("bidiSpl->UnbindDevice()");}
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   if (verbose) {util::EchoCurrentTime("done");}

   return 0;
}