////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <iostream>
#include <sstream>
#include <prntvpt.h>
#include <vector>
#include <atlimage.h>
#include "XGetopt.h"
#include "util.h"
#include "printTicketXml.h"
#include "DXP01SDK.H"

using namespace std;

struct CommandLineOptions
{
   enum DisablePrinting { All, Off, Front, Back };

   bool              checkSupplies;
   DisablePrinting   disablePrinting;
   bool              magstripe;
   short             numCopies;
   bool              portraitBack;
   bool              portraitFront;
   CString           printerName;
   bool              rotateBack;
   bool              rotateFront;
   bool              showXml;
   CString           topcoatBlockingBack;
   CString           topcoatBlockingFront;
   bool              twoPages;

   CommandLineOptions() : checkSupplies(false)
      , magstripe(false)
      , numCopies(1)
      , portraitBack(false)
      , portraitFront(false)
      , rotateBack(false)
      , rotateFront(false)
      , showXml(false)
      , twoPages(false)
      , disablePrinting(Off)
   {
   }
};

CString GetShortExeName()
{
   CString exeName;
   ::GetModuleFileName(NULL, exeName.GetBuffer(MAX_PATH), MAX_PATH);
   CPath shortExeName(exeName);
   shortExeName.StripPath();
   return shortExeName;
}

void usage()
{
   cout << endl << CT2A(GetShortExeName()) << " demonstrates print functionality of the printer and driver." << endl;
   cout << "Uses hard coded data for printing, magnetic stripe, topcoat region" << endl;
   cout << "  and print blocking region." << endl;
   cout << "Overrides the driver printing preference settings." << endl << endl;
   cout << CT2A(GetShortExeName()) << " -n <printername> [-r front | -r back] [-2]" << endl;
   cout << "  [-o frontPort | -o backPort] [ -s <number of copies>] [-m ]" << endl;
   cout << "  [-t all | -t chip | -t magJIS | -t mag2 | -t mag3 | -t custom] " << endl;
   cout << "  [-u all | -u chip | -u magJIS | -u mag2 | -u mag3 | -u custom] [-x]" << endl;
   cout << "  [-i]" << endl << endl;
   cout << "options: " << endl;
   cout << "  -n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  -r <front | back >. Rotates the card image by 180 degrees for " << endl;
   cout << "     front, back, or both sides. Use both '-r front' and '-r back'" << endl;
   cout << "     to rotate both sides 180 degrees." << endl;
   cout << "  -2 Prints a 2-sided (duplex) card. Default is front-side printing." << endl;
   cout << "  -o <frontPort | backPort >. Sets portrait orientation for a card side. " << endl;
   cout << "     Use both '-o frontPort' and '-o backPort' for portrait orientation" << endl;
   cout << "     on both sides of the card." << endl;
   cout << "     Default is landscape orientation for both card sides. " << endl;
   cout << "  -s <number of copies>. Default is 1." << endl;
   cout << "  -m Writes 3-track magnetic stripe data to backside of card using escapes. " << endl;
   cout << "     Default is no encoding." << endl;
   cout << "  -d < All | Off | Front | Back > Disable Printing. Default is Off." << endl;
   cout << "  -t <all | chip | magJIS | mag2 | mag3 | custom>" << endl;
   cout << "     Top coat and print blocking region for front of card. Use \'-t all\' to " << endl;
   cout << "     topcoat the entire card side with no print blocking. " << endl;
   cout << "     Default is the current driver setting." << endl;
   cout << "  -u <all | chip | magJIS | mag2 | mag3 | custom>" << endl;
   cout << "     Top coat and print blocking region for back of card. Use \'-u all\' to" << endl;
   cout << "     topcoat the entire card side with no print blocking." << endl;
   cout << "     Default is the current driver setting." << endl;
   cout << "  -x Display the print ticket data. Default is no display." << endl;
   cout << "  -i Check supplies before printing. Default is to not check." << endl << endl;
   cout << "  print.exe -n \"XPS Card Printer\"" << endl;
   cout << "  Prints a one-sided landscape card." << endl << endl;
   cout << "  print.exe -n \"XPS Card Printer\" -r front -r back -2 -o frontport" << endl;
   cout << "  Prints a two-sided card with both sides of card image" << endl;
   cout << "  rotated 180 degrees and with front side as portrait orientation" << endl;
   cout << "  and back side as landscape orientation." << endl << endl;
   cout << "  print -n \"XPS Card Printer\" -2 -t all -u mag3" << endl;
   cout << "  Prints a two-sided card with topcoat applied over all of side one" << endl;
   cout << "  and topcoat and printing blocked over the 3-track magnetic stripe " << endl;
   cout << "  area on the backside of the card." << endl << endl;
   ::exit(-1);
}

void CreateIStream(CComPtr <IStream>& stream) {
   stringstream exceptionText;
   HGLOBAL streamHandle(0);
   int rc = ::CreateStreamOnHGlobal(streamHandle, false, &stream);
   if (FAILED(rc)) {
      exceptionText << "::CreateStreamOnHGlobal(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
      throw runtime_error(exceptionText.str());
   }
}

void RewindIStream(CComPtr <IStream>& stream) {
   stringstream exceptionText;
   const LARGE_INTEGER BeginningOfStream = {0};
   int rc = stream->Seek(BeginningOfStream, STREAM_SEEK_SET, NULL);
   if (FAILED(rc)) {
      exceptionText << "stream->Seek(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
      throw runtime_error(exceptionText.str());
   }
}

void UpdateDevmode(const HANDLE printerHandle, const CString printerName, DEVMODE* devmode)
{
   stringstream exceptionText;

   int rc = ::DocumentProperties(
      0,
      printerHandle,
      CT2W(printerName),
      devmode, // Use the same DEVMODE for input and
      devmode, // output buffers
      DM_IN_BUFFER | DM_OUT_BUFFER);
   if (IDOK != rc) {
      const DWORD lastError = ::GetLastError();
      exceptionText << ":::DocumentProperties(" << CW2A(printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }
}

CString GetTopcoatBlockingPrintTicketXml(const CommandLineOptions& commandLineOptions)
{
   CString topcoatBlockingXml;

   // front:
   do {
      if (commandLineOptions.topcoatBlockingFront.IsEmpty()) {
         // use the current driver settings.
         break;
      }

      if (commandLineOptions.topcoatBlockingFront == L"custom") {
         // we will generate topcoat and blocking escapes for the card front.
         // escapes override the PrintTicket.
         break;
      }

      if (commandLineOptions.topcoatBlockingFront == L"all") {
         topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_All;
         break;
      }

      // we need the 'exception' markup for the remaining settings:
      topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_Except;

      if (commandLineOptions.topcoatBlockingFront == L"chip") {
         topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_ISO_7816;
         break;
      }

      if (commandLineOptions.topcoatBlockingFront == L"magjis") {
         topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_JIS;
         break;
      }

      if (commandLineOptions.topcoatBlockingFront == L"mag2") {
         topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_ISO_2Track;
         break;
      }

      if (commandLineOptions.topcoatBlockingFront == L"mag3") {
         topcoatBlockingXml += PrintTicketXml::FrontTopcoatBlockingPreset_ISO_3Track;
         break;
      }
   } while (false);

   // back:
   do {
      if (commandLineOptions.topcoatBlockingBack.IsEmpty()) {
         // use the current driver settings.
         break;
      }

      if (commandLineOptions.topcoatBlockingBack == L"custom") {
         // we will generate topcoat and blocking escapes for the card back.
         // escapes override the PrintTicket.
         break;
      }

      if (commandLineOptions.topcoatBlockingBack == L"all") {
         topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_All;
         break;
      }

      // we need the 'exception' markup for the remaining settings:
      topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_Except;

      if (commandLineOptions.topcoatBlockingBack == L"chip") {
         topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_ISO_7816;
         break;
      }

      if (commandLineOptions.topcoatBlockingBack == L"magjis") {
         topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_JIS;
         break;
      }

      if (commandLineOptions.topcoatBlockingBack == L"mag2") {
         topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_ISO_2Track;
         break;
      }

      if (commandLineOptions.topcoatBlockingBack == L"mag3") {
         topcoatBlockingXml += PrintTicketXml::BackTopcoatBlockingPreset_ISO_3Track;
         break;
      }
   } while (false);

   return topcoatBlockingXml;
}

CString GetPrintTicketXml(const CommandLineOptions& commandLineOptions)
{
   CString xml = PrintTicketXml::Prefix;

   xml += commandLineOptions.rotateFront ? PrintTicketXml::FlipFrontFlipped : PrintTicketXml::FlipFrontNone;
   xml += commandLineOptions.rotateBack ? PrintTicketXml::FlipBackFlipped : PrintTicketXml::FlipBackNone;

   switch (commandLineOptions.disablePrinting)
   {
   case CommandLineOptions::All:
      xml += PrintTicketXml::DisablePrintingAll;
      break;
   case CommandLineOptions::Off:
      xml += PrintTicketXml::DisablePrintingOff;
      break;
   case CommandLineOptions::Front:
      xml += PrintTicketXml::DisablePrintingFront;
      break;
   case CommandLineOptions::Back:
      xml += PrintTicketXml::DisablePrintingBack;
      break;
   }

   xml += GetTopcoatBlockingPrintTicketXml(commandLineOptions);
   xml += PrintTicketXml::Suffix;

   return xml;
}

void WriteCustomTopcoatBlockingEscapesFront(
   const HDC& printerDC,
   const CommandLineOptions& commandLineOptions)
{
   stringstream exceptionText;

   if (commandLineOptions.topcoatBlockingFront != L"custom")
      return;

   // a 'topcoat Add' escape will force topcoat OFF for the entire card side.

   // units are millimeters; landscape basis; top left width height:
   // a rectangle one inch down; two inches wide; 1 cm high.
   // units are millimeters; top left width height:
   CString topCoatAddEsc = L"~TA%25.4 0 50.8 10;";
   topCoatAddEsc += L"40 60 7 7?";  // add a square, lower

   int rc = ::TextOut(printerDC, 10, 10, topCoatAddEsc, topCoatAddEsc.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   // a 'blocking' escape will override the driver settings:
   const CString blockingEsc = L"~PB% 0 19 3 54;";

   rc = ::TextOut(printerDC, 40, 40, blockingEsc, blockingEsc.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }
}

void WriteCustomTopcoatBlockingEscapesBack(
   const HDC& printerDC,
   const CommandLineOptions& commandLineOptions)
{
   stringstream exceptionText;

   if (commandLineOptions.topcoatBlockingBack != L"custom")
      return;

   // a 'topcoat Remove' escape will force topcoat OFF for the entire card side.
   const CString topCoatAddEsc = L"~TA%25.4 10 50.8 20;";

   int rc = ::TextOut(printerDC, 10, 10, topCoatAddEsc, topCoatAddEsc.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   // a 'blocking' escape will override the driver settings:
   const CString blockingEsc = L"~PB% 0 23 3 54;";

   rc = ::TextOut(printerDC, 40, 40, blockingEsc, blockingEsc.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }
}

void WriteMagstripeEscapes(const HDC& printerDC)
{
   stringstream exceptionText;

   // emit some plain track 1, 2, 3 data. Assume IAT track configuration.
   const CString track1Escape(L"~1ABC 123");
   const CString track2Escape(L"~2456");
   const CString track3Escape(L"~3789");

   int rc = ::TextOut(printerDC, 0, 0, track1Escape, track1Escape.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = ::TextOut(printerDC, 0, 0, track2Escape, track2Escape.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = ::TextOut(printerDC, 0, 0, track3Escape, track3Escape.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }
}

void DrawCardBack(
   const HANDLE               printerHandle,
   const HDC&                 printerDC,
   const CommandLineOptions&  commandLineOptions,
   DEVMODE*                   devmode)
{
   stringstream exceptionText;

   // reset the devmode with proper back side orientation:
   devmode->dmOrientation  = commandLineOptions.portraitBack ? DMORIENT_PORTRAIT : DMORIENT_LANDSCAPE;
   UpdateDevmode(printerHandle, commandLineOptions.printerName, devmode);
   if (NULL == ::ResetDC(printerDC, devmode)) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "::ResetDC(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   int rc = ::StartPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "StartPage(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CString text;
   text.Format(L"%s back on", GetShortExeName());
   rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   text.Format(L"%s", commandLineOptions.printerName);
   rc = ::TextOut(printerDC, 50, 100, text, text.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CImage kBitmap;
   CString bitmapName = L"House_1bpp.png";
   rc = kBitmap.Load(bitmapName);
   if (FAILED(rc)) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "Load(" << CW2A(bitmapName) << ") " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = kBitmap.Draw(printerDC, 50, 150);
   if (!rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "kBitmap.Draw(): " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CImage colorBitmap;
   bitmapName = L"color.jpg";
   rc = colorBitmap.Load(bitmapName);
   if (FAILED(rc)) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "Load(" << CW2A(bitmapName) << ") " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = colorBitmap.Draw(printerDC, 50, 300);
   if (!rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "colorBitmap.Draw(): " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   WriteCustomTopcoatBlockingEscapesBack(printerDC, commandLineOptions);

   rc = ::EndPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "EndPage(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }
}

void DrawCardFront(const HDC& printerDC, const CommandLineOptions& commandLineOptions)
{
   stringstream exceptionText;

   int rc = ::StartPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "StartPage(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   if (commandLineOptions.magstripe) {
      WriteMagstripeEscapes(printerDC);
   }

   CString text;
   text.Format(L"%s front on", GetShortExeName());
   rc = ::TextOut(printerDC, 50, 50, text, text.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   text.Format(L"%s", commandLineOptions.printerName);
   rc = ::TextOut(printerDC, 50, 100, text, text.GetLength());
   if (0 == rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "TextOut(): " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CImage kBitmap;
   CString bitmapName = L"House_1bpp.png";
   rc = kBitmap.Load(bitmapName);
   if (FAILED(rc)) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "Load(" << CW2A(bitmapName) << ") " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = kBitmap.Draw(printerDC, 50, 150);
   if (!rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "kBitmap.Draw(): " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   CImage colorBitmap;
   bitmapName = L"color.jpg";
   rc = colorBitmap.Load(bitmapName);
   if (FAILED(rc)) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "Load(" << CW2A(bitmapName) << ") " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = colorBitmap.Draw(printerDC, 50, 300);
   if (!rc) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "colorBitmap.Draw(): " << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   WriteCustomTopcoatBlockingEscapesFront(printerDC, commandLineOptions);

   rc = ::EndPage(printerDC);
   if (rc <= 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "EndPage(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }
}

int CreatePrintJob(
   const HANDLE               printerHandle,
   const CommandLineOptions&  commandLineOptions,
   DEVMODE*                   devmode)
{
   stringstream exceptionText;

   HDC printerDC = ::CreateDC(
      L"WINSPOOL",
      devmode->dmDeviceName,
      NULL,
      devmode);
   if (NULL == printerDC) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "CreateDC(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   DOCINFO docInfo = {0};
   docInfo.cbSize = sizeof DOCINFO;
   docInfo.lpszDocName = L"xps printer sdk print sample document";
   int rc = ::StartDoc(printerDC, &docInfo);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "StartDoc(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   const DWORD windowsJobID = rc; // needed for job completion

   DrawCardFront(printerDC, commandLineOptions);

   if (commandLineOptions.twoPages)
   {
      DrawCardBack(printerHandle, printerDC, commandLineOptions, devmode);
   }

   rc = ::EndDoc(printerDC);
   if (rc < 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "EndDoc(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   rc = ::DeleteDC(printerDC);
   if (rc == 0) {
      const DWORD lastErr = ::GetLastError();
      exceptionText << "DeleteDC(): " << " 0x" << hex << lastErr << " " << CW2A(util::Win32ErrorString(lastErr));
      throw runtime_error(exceptionText.str());
   }

   return windowsJobID;
}

int PrintTextAndGraphics(const CommandLineOptions& commandLineOptions)
{
   stringstream exceptionText;

   // for the given printer get the maxSchemaVersion needed for the
   // PTOpenProvider() call:
   DWORD maxSchemaVersion(0);
   int rc = ::PTQuerySchemaVersionSupport(commandLineOptions.printerName, &maxSchemaVersion);
   if (FAILED(rc)) {
      const DWORD lastError = ::GetLastError();
      exceptionText << "::PTQuerySchemaVersionSupport("
         << CW2A(commandLineOptions.printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   // the the handle to the printticket printTicketProviderHandle.
   HPTPROVIDER printTicketProviderHandle(0);
   rc = ::PTOpenProvider(commandLineOptions.printerName, maxSchemaVersion, &printTicketProviderHandle);
   if (FAILED(rc)) {
      const DWORD lastError = ::GetLastError();
      exceptionText << "::DPTOpenProvider(" << CW2A(commandLineOptions.printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   // get a printer handle:
   HANDLE printerHandle(0);
   rc = ::OpenPrinter(CT2W(commandLineOptions.printerName), &printerHandle, NULL);
   if (!rc) {
      const DWORD lastError = ::GetLastError();
      exceptionText << "::OpenPrinter(" << CW2A(commandLineOptions.printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   // get a DEVMODE. we need to get the buffer size first.
   rc = ::DocumentProperties(
      0,
      printerHandle,
      CT2W(commandLineOptions.printerName),
      NULL,
      NULL,
      0);
   if (0 > rc) {
      const DWORD lastError = ::GetLastError();
      exceptionText << ":::DocumentProperties(" << CW2A(commandLineOptions.printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   ULONG devmode_size = rc;

   // this buffer will nicely self-destruct:
   vector <byte> devmodeBytes(devmode_size, 0);
   DEVMODE* devmode = (DEVMODE*) (&devmodeBytes[0]);

   rc = ::DocumentProperties(
      NULL,
      printerHandle,
      CT2W(commandLineOptions.printerName),
      devmode,
      NULL,
      DM_OUT_BUFFER);
   if (rc < 0) {
      const DWORD lastError = ::GetLastError();
      exceptionText << ":::DocumentProperties(" << CW2A(commandLineOptions.printerName)
         << ") " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   // use the DEVMODE for standard devmode items:
   devmode->dmCopies      = commandLineOptions.numCopies;
   devmode->dmDuplex      = commandLineOptions.twoPages ? DMDUP_HORIZONTAL : DMDUP_SIMPLEX;
   devmode->dmOrientation = commandLineOptions.portraitFront ? DMORIENT_PORTRAIT : DMORIENT_LANDSCAPE;

   UpdateDevmode(printerHandle, commandLineOptions.printerName, devmode);

   CComPtr <IStream> printTicketStream;
   CreateIStream(printTicketStream);

   rc = ::PTConvertDevModeToPrintTicket(
      printTicketProviderHandle,
      devmode_size,
      devmode,
      kPTJobScope,
      printTicketStream);
   if (FAILED(rc)) {
      exceptionText << "::PTConvertDevModeToPrintTicket(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
      throw runtime_error(exceptionText.str());
   }

   rc = printTicketStream->Commit(0);
   if (FAILED(rc)) {
      exceptionText << "printTicketStream->Commit(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
      throw runtime_error(exceptionText.str());
   }

   /////////////////////////////////////////////////////////////////////////////
   // construct the entire print ticket now, regardless of whether or not we
   // are printing one or twp sides.
   /////////////////////////////////////////////////////////////////////////////
   CString printTicketXml = GetPrintTicketXml(commandLineOptions);

   CComPtr <IStream> deltaTicketStream;
   CreateIStream(deltaTicketStream);

   ULONG deltaTicketStreamBytesWritten(0);
   rc = deltaTicketStream->Write(
      CW2A(printTicketXml.GetBuffer()),
      printTicketXml.GetLength(),
      &deltaTicketStreamBytesWritten);
   if (FAILED(rc)) {
      exceptionText << "deltaTicketStream->Write(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
      throw runtime_error(exceptionText.str());
   }

   CComPtr <IStream> resultTicketStream;
   CreateIStream(resultTicketStream);
   RewindIStream(printTicketStream);
   RewindIStream(deltaTicketStream);

   CComBSTR printTicketErrorMessage;
   rc = ::PTMergeAndValidatePrintTicket(
      printTicketProviderHandle,
      printTicketStream,
      deltaTicketStream,
      kPTJobScope,
      resultTicketStream,
      &printTicketErrorMessage);
   if (S_PT_NO_CONFLICT != rc) {
      exceptionText << "::PTMergeAndValidatePrintTicket(): " << rc << " "
         << CT2A(util::Win32ErrorString(rc)) << " " << CW2T(printTicketErrorMessage);
      throw runtime_error(exceptionText.str());
   }

   if (commandLineOptions.showXml)
   {
      // echo the resultant print ticket:

      RewindIStream(resultTicketStream);
      STATSTG statStorage = {0};

      rc = resultTicketStream->Stat(&statStorage, STATFLAG_DEFAULT);
      if (FAILED(rc)) {
         exceptionText << "resultTicketStream->Stat(): " << rc << CT2A(util::Win32ErrorString(rc));
         throw runtime_error(exceptionText.str());
      }

      const int ticketSize = statStorage.cbSize.LowPart;

      vector <byte> printTicketBytes(ticketSize, 0);
      ULONG bytesWritten(0);
      rc = resultTicketStream->Read(&printTicketBytes[0], ticketSize, &bytesWritten);
      if (FAILED(rc)) {
         exceptionText << "resultTicketStream->Write(): " << rc << " " << CT2A(util::Win32ErrorString(rc));
         throw runtime_error(exceptionText.str());
      }

      CString printTicketXml = (char*) printTicketBytes.data();
      cout << endl << CT2A(printTicketXml) << endl;
   }

   RewindIStream(resultTicketStream);

   // convert the merged printticket to a DEVMODE. we'll use the
   // DEVMODE as a parameter to the CreateDC() call.
   rc = ::PTConvertPrintTicketToDevMode(
      printTicketProviderHandle,
      resultTicketStream,
      kUserDefaultDevmode,
      kPTJobScope,
      &devmode_size,
      &devmode,
      &printTicketErrorMessage);
   if (FAILED(rc)) {
      exceptionText << "::PTConvertPrintTicketToDevMode(): " << rc << " "
         << CT2A(util::Win32ErrorString(rc)) << " " << CW2T(printTicketErrorMessage);
      throw runtime_error(exceptionText.str());
   }

   const int windowsJobID = CreatePrintJob(printerHandle, commandLineOptions, devmode);

   rc = ::PTCloseProvider(printTicketProviderHandle);
   if (FAILED(rc)) {
      const DWORD lastError = ::GetLastError();
      exceptionText << "::PTCloseProvider() " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   rc = ::ClosePrinter(printerHandle);
   if (FAILED(rc)) {
      const DWORD lastError = ::GetLastError();
      exceptionText << "::ClosePrinter() " << lastError << " " << CT2A(util::Win32ErrorString(lastError));
      throw runtime_error(exceptionText.str());
   }

   // free the devmode from the PTConvertPrintTicketToDevMode() call:
   rc = ::PTReleaseMemory(devmode);

   return windowsJobID;
}

void CancelCard(CComPtr <IBidiSpl>& bidiSpl, const util::PrinterStatusValues& printerStatusValues)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   PBYTE xmlBytes = (PBYTE) ::CoTaskMemAlloc(xmlBytesLength);

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   ::CoTaskMemFree(xmlBytes);
}

////////////////////////////////////////////////////////////////////////////////
// CheckSupplies()
//
// issue an IBidi StartJob with xml markup to start an IBidi job, just to have
// the printer check for supplies. If the printer is out of supplies, throw an
// exception.
//
// if not problem, return the printJobID.
////////////////////////////////////////////////////////////////////////////////
int CheckSupplies(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // first, get the printer options. we need this to construct XML passed to
   // the StartJob().
   const CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
   const util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   const short checkPrintRibbon =
      printerOptionsValues._printHead == TEXT("Installed") ? 1 : 0;

   const short checkEmbossSupplies =
      printerOptionsValues._moduleEmbosser == "None" ? 0 : 1;

   CString startjobXml;
   startjobXml.Format(
      dxp01sdk::STARTJOB_XML,
      TEXT(""),   // hopperNumber
      checkPrintRibbon,
      checkEmbossSupplies);

   const ULONG startjobDataLength = startjobXml.GetLength() * sizeof(TCHAR);

   // Get memory that can be used with IBidiSpl:
   CComAllocator XMLBytesAllocator; // automatically frees COM memory
   PBYTE XMLBytes = (PBYTE) XMLBytesAllocator.Allocate(startjobDataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(XMLBytes, startjobDataLength, startjobXml.GetBuffer(), startjobDataLength);

   // pass the startjob input data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, XMLBytes, startjobDataLength);
   if (FAILED(hr)) {
      exceptionText << "SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      // the StartJob failed; send a Cancel(); do NOT send an EndJob():
      CancelCard(bidiSpl, printerStatusValues);
      exceptionText << "startjob(): " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }

   return printerStatusValues._printerJobID;
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

void ValidateCommandLineOptions(CommandLineOptions& commandLineOptions)
{
   if (commandLineOptions.printerName.IsEmpty()) {usage();}

   if (commandLineOptions.numCopies <= 0) {
      cout << "bad number of copies: " << commandLineOptions.numCopies << endl; ::exit(-1);
   }

   if (
      commandLineOptions.topcoatBlockingFront != L""       &&
      commandLineOptions.topcoatBlockingFront != L"all"    &&
      commandLineOptions.topcoatBlockingFront != L"chip"   &&
      commandLineOptions.topcoatBlockingFront != L"magjis" &&
      commandLineOptions.topcoatBlockingFront != L"mag2"   &&
      commandLineOptions.topcoatBlockingFront != L"mag3"   &&
      commandLineOptions.topcoatBlockingFront != L"custom")
   {
      cout << "invalid front topcoat / blocking option: " << CW2A(commandLineOptions.topcoatBlockingFront) << endl;
      ::exit(-1);
   }

   if (
      commandLineOptions.topcoatBlockingBack != L""       &&
      commandLineOptions.topcoatBlockingBack != L"all"    &&
      commandLineOptions.topcoatBlockingBack != L"chip"   &&
      commandLineOptions.topcoatBlockingBack != L"magjis" &&
      commandLineOptions.topcoatBlockingBack != L"mag2"   &&
      commandLineOptions.topcoatBlockingBack != L"mag3"   &&
      commandLineOptions.topcoatBlockingBack != L"custom")
   {
      cout << "invalid back topcoat / blocking option: " << CW2A(commandLineOptions.topcoatBlockingBack) << endl;
      ::exit(-1);
   }
}

CommandLineOptions GetCommandlineOptions(const int argc, _TCHAR* argv[])
{
   CommandLineOptions commandLineOptions;

   int c(0);
   while ((c = getopt(argc, argv, _T("n:r:o:xs:2md:t:u:i"))) != EOF) {
      switch (c) {
      case L'n': commandLineOptions.printerName = optarg; break;
      case L'r':
         if (L"front" == CString(optarg).MakeLower()) {commandLineOptions.rotateFront = true; break;}
         if (L"back"  == CString(optarg).MakeLower()) {commandLineOptions.rotateBack  = true; break;}
         break;
      case L'o':
         if (L"frontport" == CString(optarg).MakeLower()) {commandLineOptions.portraitFront = true; break;}
         if (L"backport"  == CString(optarg).MakeLower()) {commandLineOptions.portraitBack  = true; break;}
         break;
      case L'x': commandLineOptions.showXml = true; break;
      case L's': commandLineOptions.numCopies = (short) ::_wtoi(optarg); break;
      case L'2': commandLineOptions.twoPages = true; break;
      case L'm': commandLineOptions.magstripe = true; break;

      case L'd':
         {
            if (L"all"   == CString(optarg).MakeLower()) {commandLineOptions.disablePrinting = CommandLineOptions::All;   break;}
            if (L"off"   == CString(optarg).MakeLower()) {commandLineOptions.disablePrinting = CommandLineOptions::Off;   break;}
            if (L"front" == CString(optarg).MakeLower()) {commandLineOptions.disablePrinting = CommandLineOptions::Front; break;}
            if (L"back"  == CString(optarg).MakeLower()) {commandLineOptions.disablePrinting = CommandLineOptions::Back;  break;}
            usage();
         }

      case L't': commandLineOptions.topcoatBlockingFront = CString(optarg).MakeLower(); break;
      case L'u': commandLineOptions.topcoatBlockingBack  = CString(optarg).MakeLower(); break;
      case L'i': commandLineOptions.checkSupplies = true; break;
      default:
         usage();
      }
   }

   return commandLineOptions;
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   CommandLineOptions commandLineOptions = GetCommandlineOptions(argc, argv);
   ValidateCommandLineOptions(commandLineOptions);

   CComPtr <IBidiSpl> bidiSpl;

   try {
      stringstream exceptionText;
      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // create an IBidiSpl COM smartpointer instance, and bind it to the
      // printer name. We use it for all subsequent call using the IBidi interface.
      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(commandLineOptions.printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice() error " << " 0x" << hex << hr << " '" << CT2A(commandLineOptions.printerName) << "' " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      const util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);
      if ("Ready" != printerOptionsValues._printerStatus) {
         exceptionText << "printer " << CT2A(commandLineOptions.printerName) << " is not ready.";
         string dlftest = exceptionText.str();
         throw runtime_error(dlftest);
      }

      int printJobID(0);

      if (commandLineOptions.checkSupplies) {
         printJobID = CheckSupplies(bidiSpl);
      }

      const int windowsJobID = PrintTextAndGraphics(commandLineOptions);

      if (0 != printJobID) {
         // we issued an IBidi StartJob() to check for supplies. We need to
         // wait for all the print data to be spooled.
         HANDLE printerHandle(0);
         BOOL brc = ::OpenPrinter(commandLineOptions.printerName.GetBuffer(), &printerHandle, NULL);
         if (!brc) {
            const DWORD lastErr = ::GetLastError();
            exceptionText << "OpenPrinter():" << " 0x" << hex << lastErr << " " << CT2A(util::Win32ErrorString(lastErr));
            throw runtime_error(exceptionText.str());
         }

         util::WaitUntilJobSpooled(printerHandle, windowsJobID); 
         ::ClosePrinter(printerHandle);
      }

      EndJob(bidiSpl);
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}
