////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <iostream>
#include <sstream>
#include <vector>
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

enum Commands
{
   undefined,
   cancelAllJobs
};

struct CommandLineOptions
{
   CommandLineOptions() : verbose(false)
   , command(undefined)
   {}

   CString  printerName;
   Commands command;
   bool     verbose;
};

CString GetShortExeName()
{
   CString exeName;
   ::GetModuleFileName(NULL, exeName.GetBuffer(MAX_PATH), MAX_PATH);
   CPath shortExeName(exeName);
   shortExeName.StripPath();
   return shortExeName;
}

void usage()
{
   cout << endl << CT2A(GetShortExeName()) << " controls the printer through the printer driver." << endl << endl;
   cout << CT2A(GetShortExeName()) << " -n <printername> <command>" << endl;
   cout << "  -n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  -c cancel all jobs: Send a command to the printer to clear any print job" << endl;
   cout << "                    from the printer. This does not delete spooled Windows" << endl;
   cout << "                    print jobs." << endl;
   cout << "  -v Verbose. Optional." << endl << endl;
   ::exit(-1);
}

void ValidateCommandLineOptions(CommandLineOptions& commandLineOptions)
{
   if (commandLineOptions.printerName.IsEmpty()) {usage();}
   if (undefined == commandLineOptions.command) {usage();}
}

CommandLineOptions GetCommandlineOptions(const int argc, _TCHAR* argv[])
{
   CommandLineOptions commandLineOptions;
   int c(0);
   while ((c = getopt(argc, argv, _T("n:cv"))) != EOF) {
      switch (c) {
      case L'n':
         commandLineOptions.printerName = optarg;
         break;
      case L'c':
         commandLineOptions.command = cancelAllJobs;
         break;
      case L'v':
         commandLineOptions.verbose = true;
         break;
      default:
         usage();
      }
   }
   return commandLineOptions;
}

void DisplayXML(CString title, CString xml)
{
   cout << CW2A(title) << ":" << endl;
   cout << CW2A(xml) << endl;
   cout << endl;
}

////////////////////////////////////////////////////////////////////////////////
// CancelAllJobs()
//
// using the IBidiSpl interface, issue a 'cancel all jobs' command. This is
// simply a dxp01sdk::Cancel with both printerJobID and errorCode set to zero.
//
////////////////////////////////////////////////////////////////////////////////
void CancelAllJobs(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // use zero for jobID and errorCode; this will cancel all cards in printer.
   CString actionXML = util::FormatPrinterActionXML(dxp01sdk::Cancel, 0, 0);

   if (verbose) DisplayXML("printer action", actionXML);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   CommandLineOptions commandLineOptions = GetCommandlineOptions(argc, argv);

   ValidateCommandLineOptions(commandLineOptions);

   CComPtr <IBidiSpl> bidiSpl;
   try {
      stringstream exceptionText;

      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // create an IBidiSpl COM smartpointer instance, and bind it to the
      // printer name. We use it for all subsequent call using the IBidi interface.
      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(commandLineOptions.printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(commandLineOptions.printerName) << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      if (cancelAllJobs == commandLineOptions.command) {
         CancelAllJobs(bidiSpl, commandLineOptions.verbose);
         cout << "all jobs canceled for '" << CT2A(commandLineOptions.printerName) << "'" << endl;
      }
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}
