////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps driver sdk loosely-coupled c++ smartcard sample
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <bidispl.h>
#include <WinSpool.h>
#include <iostream>
#include <sstream>
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

void usage(CString longExeName)
{
   CPath shortExeName(longExeName);
   shortExeName.StripPath();

   cout << endl;
   cout << CT2A(shortExeName) << " demonstrates interactive mode parking a card in the smart card" << endl;
   cout << "station, moving the card from the station, and options to print" << endl;
   cout << "and poll for job completion." << endl;
   cout << ">Uses hard-coded data for printing." << endl << endl;
   cout << CT2A(shortExeName) << " -n <printername> [-p] [-c]" << endl << endl;
   cout << "options:" << endl;
   cout << "  " << "-n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  " << "-p Print simple black text on one or both sides of the card"  << endl; 
   cout << "  " << "   depending on the printer used."  << endl; 
   cout << "  " << "-c Poll for job status. " << endl;
   shortExeName.RemoveExtension();
   cout << endl << CT2A(shortExeName) << " -n \"XPS Card Printer\"" << endl;
   cout << "Parks a card in the printer smart card station, asks you to continue or" << endl;
   cout << "reject and then does what you requested." << endl << endl;
   cout << endl << CT2A(shortExeName) << " -n \"XPS Card Printer\" -p -c" << endl;
   cout << "Parks a card in printer smart card station, moves the card from the" << endl;
   cout << "station, prints black text on one or both sides of the card," << endl;
   cout << "and polls and displays job status." << endl << endl;
   exit(0);
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

void CancelCard(CComPtr <IBidiSpl>& bidiSpl, const util::PrinterStatusValues& printerStatusValues)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   PBYTE xmlBytes = (PBYTE) ::CoTaskMemAlloc(xmlBytesLength);

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   ::CoTaskMemFree(xmlBytes);
}

void StartJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {

      // the StartJob failed; send a Cancel(); do NOT send an EndJob():
      CancelCard(bidiSpl, printerStatusValues);

      exceptionText << "startjob():" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }
}

util::PrinterStatusValues ParkCard(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::SMARTCARD_PARK);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::XPS_CARDPRINTER_SDK_SCHEMA_SMARTCARD_PARK):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) park:" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   HRESULT parkCardResultCode(S_OK);
   hr = bidiRequest->GetResult(&parkCardResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(park):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(parkCardResultCode)) {
      exceptionText << "parkCardResultCode:" << " 0x" << hex << parkCardResultCode << " " << util::Win32ErrorString(parkCardResultCode).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues); // the ParkCard failed; send a Cancel(); 
      exceptionText << "startjob(): " << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   return printerStatusValues;
}

void ResumeCard(const int printerJobID, CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call:
   CString actionXML = util::FormatPrinterActionXML(dxp01sdk::Resume, printerJobID, 0);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR; // UNICODE

   // allocate IBidiSpl compatible memory:
   PBYTE xmlBytes = (PBYTE) ::CoTaskMemAlloc(xmlBytesLength);

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::XPS_CARDPRINTER_SCHEMA_PRINTER_ACTION):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   ::CoTaskMemFree(xmlBytes);
}


////////////////////////////////////////////////////////////////////////////////
// PersonalizeSmartcard()
////////////////////////////////////////////////////////////////////////////////
bool PersonalizeSmartcard()
{
   CString caption(L"A card has been parked in smartcard module");

   CString message(
      L"This is where smartcard personalization happens.\r\n\r\n"
      L"Press 'Yes' to simulate a successful smartcard personalization.\r\n\r\n"
      L"Press 'No' to simulate a failed smartcard personalization.");

   int rc = ::MessageBox(NULL, message, caption, MB_YESNO);

   return IDYES == rc;
}

////////////////////////////////////////////////////////////////////////////////
// PollForJobCompletion()
//
// repeatedly check job status until the job is complete.
////////////////////////////////////////////////////////////////////////////////
void PollForJobCompletion(CComPtr <IBidiSpl>& bidiSpl, const int printerJobID)
{
   for (; true ;)
   {
      const CString jobStatusXML = util::GetJobStatusXML(bidiSpl, printerJobID);
      const util::JobStatusValues jobStatusValues = util::ParseJobStatusXML(jobStatusXML);

      cout << "Printer job id " << printerJobID << " " << CW2A(jobStatusValues._jobState) << endl;

      if (dxp01sdk::JOB_ACTIVE != jobStatusValues._jobState)
      {
         return;
      }

      ::Sleep(1000); // milliseconds
   }
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   CString printerName;
   int c(0);
   bool print(false);
   bool pollForJobCompletion(false);

   while ((c = getopt(argc, argv, _T("n:pc"))) != EOF) {
      switch (c) {
      case L'n':
         printerName = optarg; break;
      case L'p':
         print = true; break;
      case L'c':
         pollForJobCompletion = true; break;
      default:
         usage(argv[0]);
      }
   }

   if (printerName.IsEmpty()) {
      usage(argv[0]);
   }

   stringstream exceptionText;
   CComPtr <IBidiSpl> bidiSpl;

   try {
      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(printerName) << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      // see if the printer is in the Printer_Ready state:
      CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);

      if ("Ready" != printerOptionsValues._printerStatus) {
         exceptionText << "printer " << CT2A(printerName) << " is not ready.";
         throw runtime_error(exceptionText.str());
      }

      StartJob(bidiSpl);

      // park the card in the smartcard reader:
      util::PrinterStatusValues printerStatusValues = ParkCard(bidiSpl);

      // the card has been parked in the smartcard reader:
      const bool personalizedOK = PersonalizeSmartcard();

      if (!personalizedOK) {
         cout << "smartcard personalization failed." << endl;
         CancelCard(bidiSpl, printerStatusValues);
      }
      else {
         cout << "smartcard personalization succeeded." << endl;

         ResumeCard(printerStatusValues._printerJobID, bidiSpl);

         if (print) {
            // do some simple text and graphics printing.

            const bool doTwoSided = 
               printerOptionsValues._optionDuplex == "Auto" ||
               printerOptionsValues._optionDuplex == "Manual";

            // Important: util::PrintTextmethod also waits until driver gets all the print data.
            util::PrintText(printerName, doTwoSided);
         }
      }

      EndJob(bidiSpl);

      if (pollForJobCompletion) {
         PollForJobCompletion(bidiSpl, printerStatusValues._printerJobID);
      }
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}