////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// XPS Printer Driver SDK: Single-wire smartcard personalization c++ sample.
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <bidispl.h>
#include <WinSpool.h>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include "DXP01SDK_SCARD.H"
#include "DXP01SDK_SCARD_STRINGS.H"
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

void usage(CString longExeName)
{
   CPath shortExeName(longExeName);
   shortExeName.StripPath();
   cout << endl;
   cout << CT2A(shortExeName) << " demonstrates smartcard personalization using the"  << endl;
   cout << "single-wire embedded reader. Once staged, the app connects to a contact" << endl;
   cout << "and then a contactless chip. For each, an ATR is issued and some data bytes" << endl;
   cout << "are sent and received." << endl <<endl;
   cout << "Optionally prints, encodes a magstripe, and waits for job completion." << endl << endl;
   cout << "Uses hard-coded data for printing." << endl << endl;
   cout << CT2A(shortExeName) << " -n <printername> [-p] [-c] [-i] [-m] [-v]" << endl << endl;
   cout << "options:" << endl;
   cout << "  " << "-n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  " << "-p Print simple black text on one or both sides of the card"  << endl; 
   cout << "  " << "   depending on the printer used. Optional."  << endl; 
   cout << "  " << "-i <input hopper>. Optional. Defaults to \"1\"."  << endl; 
   cout << "  " << "-c Poll for job status. Optional." << endl;
   cout << "  " << "-m Encode a mag stripe. Optional." << endl;
   cout << "  " << "-v verbose mode with timing information. Optional." << endl << endl;
   shortExeName.RemoveExtension();
   exit(0);
}

void CancelCard(
   CComPtr <IBidiSpl>& bidiSpl,
   const util::PrinterStatusValues& printerStatusValues,
   const bool verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("CancelCard(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("CancelCard(): return");}
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EndJob(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EndJob(): return");}
}

void EncodeMagstripe(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // Hardcoded XML to encode all 3 tracks in IAT mode.
   // track 1 = "TRACK1", track 2 = "1122", track 3 = "321"
   CString trackDataXML = L"<?xml version=\"1.0\" encoding=\"utf-8\"?>"
      L"<magstripe>"
      L"<track number=\"1\"><base64Data>VFJBQ0sx</base64Data></track>"
      L"<track number=\"2\"><base64Data>MTEyMg==</base64Data></track>"
      L"<track number=\"3\"><base64Data>MzIx</base64Data></track>"
      L"</magstripe>";

   // Data is the length of the string multiplied by size of wide char
   ULONG dataLength = trackDataXML.GetLength() * sizeof WCHAR;

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE trackDataXMLBytes = (PBYTE) XMLBytesAllocator.Allocate(dataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(trackDataXMLBytes, dataLength, trackDataXML.GetBuffer(), dataLength);

   // Setup the IBidiSpl request object using IBidiRequest ...
   // First, we are going to encode a magstripe.
   hr = bidiRequest->SetSchema(dxp01sdk::MAGSTRIPE_ENCODE);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::MAGSTRIPE_ENCODE):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // pass the magstripe data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, trackDataXMLBytes, dataLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EncodeMagStripe(): calling SendRecv()");}

   // Send the command to the printer.
   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiSpl->SendRecv(BIDI_ACTION_SET):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("EncodeMagStripe(): SendRecv() return");}

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues, verbose);
      exceptionText << "EncodeMagStripe():" << " 0x" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }
}

void StartJob(
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           verbose,
   const wstring        inputHopperName = L"1")
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString startjobXml;
   startjobXml.Format(
      dxp01sdk::STARTJOB_XML,
      inputHopperName.c_str(),
      0,    // zero means "do not check print ribbon"
      0);   // zero means "do not check emboss supplies"

   const ULONG startjobDataLength = startjobXml.GetLength() * sizeof(TCHAR);

   // Get memory that can be used with IBidiSpl:
   CComAllocator XMLBytesAllocator; // automatically frees COM memory
   PBYTE XMLBytes = (PBYTE) XMLBytesAllocator.Allocate(startjobDataLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...but not the terminating null:
   ::memcpy_s(XMLBytes, startjobDataLength, startjobXml.GetBuffer(), startjobDataLength);

   // pass the startjob input data:
   hr = bidiRequest->SetInputData(BIDI_BLOB, XMLBytes, startjobDataLength);
   if (FAILED(hr)) {
      exceptionText << "SetInputData():" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("StartJob(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("StartJob(): bidiSpl->SendRecv() return");}

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      // the StartJob failed; send a Cancel(); do NOT send an EndJob():
      CancelCard(bidiSpl, printerStatusValues, verbose);
      exceptionText << "startjob(): " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("StartJob(): return");}
}

util::PrinterStatusValues ParkCard(
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::SMARTCARD_PARK);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::XPS_CARDPRINTER_SDK_SCHEMA_SMARTCARD_PARK):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ParkCard(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) park:" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   HRESULT parkCardResultCode(S_OK);
   hr = bidiRequest->GetResult(&parkCardResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(park):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(parkCardResultCode)) {
      exceptionText << "parkCardResultCode:" << " 0x" << hex << parkCardResultCode << " " << util::Win32ErrorString(parkCardResultCode).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues, verbose);
      exceptionText << "startjob(): " << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString);
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ParkCard(): return");}

   return printerStatusValues;
}

void ResumeCard(
   const int            printerJobID,
   CComPtr <IBidiSpl>&  bidiSpl,
   const bool           verbose)
{
   stringstream exceptionText;
   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call:
   CString actionXML = util::FormatPrinterActionXML(dxp01sdk::Resume, printerJobID, 0);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR; // UNICODE

   // IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::XPS_CARDPRINTER_SCHEMA_PRINTER_ACTION):" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData():" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ResumeCard(): calling bidiSpl->SendRecv(BIDI_ACTION_SET)");}

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)){
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml:" << " 0x" << hex << hr << " " << util::Win32ErrorString(hr).GetBuffer();
      throw runtime_error(exceptionText.str());
   }

   if (verbose) {util::EchoCurrentTime("ResumeCard(): return");}
}

////////////////////////////////////////////////////////////////////////////////
// PollForJobCompletion()
//
// repeatedly check job status until the job is complete.
////////////////////////////////////////////////////////////////////////////////
void PollForJobCompletion(
   CComPtr <IBidiSpl>&  bidiSpl,
   const int            printerJobID,
   const bool           verbose)
{
   for (; true ;)
   {
      const CString jobStatusXML = util::GetJobStatusXML(bidiSpl, printerJobID);
      const util::JobStatusValues jobStatusValues = util::ParseJobStatusXML(jobStatusXML);

      CString message;
      message.Format(TEXT("Printer jobID: %d %s"), printerJobID, jobStatusValues._jobState);

      if (verbose) {util::EchoCurrentTime(message);}
      else {cout << CT2A(message) << endl;}

      if (dxp01sdk::JOB_ACTIVE != jobStatusValues._jobState) {
         return;
      }
      ::Sleep(1000); // milliseconds
   }
}

////////////////////////////////////////////////////////////////////////////////
// SCardResultMessage()
// format a nice message for display.
// SCARD errors are declared in WinError.H
////////////////////////////////////////////////////////////////////////////////
string SCardResultMessage(const long scardResult, const string message)
{
   stringstream exceptionText;
   exceptionText << message
      << " result: "
      << scardResult
      << "; "
      << CT2A(util::Win32ErrorString(scardResult));
   return exceptionText.str();
}

////////////////////////////////////////////////////////////////////////////////
// PersonalizeChip()
//
// talk to the chip (contact or contactless)
////////////////////////////////////////////////////////////////////////////////
void PersonalizeChip(dxp01sdk::SCard& scard, const bool verbose)
{
   stringstream exceptionText;

   vector <DWORD> states;
   vector <byte> ATRBytes;
   DWORD protocol(SCARD_PROTOCOL_UNDEFINED);

   if (verbose) {util::EchoCurrentTime("calling SCardStatus()");}

   long scardResult = scard.SCardStatus(states, protocol, ATRBytes);
   cout << SCardResultMessage(scardResult, "SCardStatus()") << endl;

   if (scardResult != SCARD_S_SUCCESS) {
      // unable to get a valid SCardStatus() response; we're done.
      return;
   }

   cout << "SCardStatus() states: ";
   for (size_t index = 0; index < states.size(); index++) {
      wcout << states[index] << " " << dxp01sdk::SCard::StringFromState(states[index]) << " ";
   } cout << endl;

   const vector <DWORD>::iterator cardAbsent = find(states.begin(), states.end(), SCARD_ABSENT);
   if (cardAbsent != states.end()) {
      // one of the states is SCARD_ABSENT; we're done.
      return;
   }

   cout << "SCardStatus() ATRBytes (hex): ";
   for (size_t index = 0; index < ATRBytes.size(); index++) {
      cout << setw(2) << setfill('0') << uppercase << hex << int(ATRBytes[index]) << " ";
   }
   cout << dec << endl;

   // try all these 'get attribute' items:
   DWORD attributesToTryArray[] = {
      SCARD_ATTR_VENDOR_NAME,
      SCARD_ATTR_VENDOR_IFD_SERIAL_NO,
      SCARD_ATTR_VENDOR_IFD_TYPE,
      SCARD_ATTR_VENDOR_IFD_VERSION
   };
   vector <DWORD> attributesToTry(attributesToTryArray, attributesToTryArray + ARRAYSIZE(attributesToTryArray));

   for (size_t attribIndex = 0; attribIndex < attributesToTry.size(); attribIndex++)
   {
      vector <byte> scardAttributeBytes;

      if (verbose) {
         CString s;
         s.Format(TEXT("calling SCardGetAttrib(%d)"), attributesToTry[attribIndex]);
         util::EchoCurrentTime(s);
      }

      scardResult = scard.SCardGetAttrib(attributesToTry[attribIndex], scardAttributeBytes);
      if (SCARD_S_SUCCESS != scardResult)
         continue;

      CString msg;
      msg.Format(L"SCardGetAttrib(%ls)", dxp01sdk::SCard::StringFromAttr(attributesToTry[attribIndex]).c_str());
      cout << SCardResultMessage(scardResult, string(CT2A(msg))) << endl;

      switch (attributesToTry[attribIndex])
      {
      case SCARD_ATTR_VENDOR_IFD_VERSION:
         {
            // vendor-supplied interface device version: DWORD in the form 0xMMmmbbbb where
            //      MM = major version;
            //      mm = minor version; and
            //      bbbb = build number:
            const size_t byteCount = scardAttributeBytes.size();
            const short minorVersion = byteCount > 0 ? scardAttributeBytes[0] : 0;
            const short majorVersion = byteCount > 1 ? scardAttributeBytes[1] : 0;
            short buildNumber = 0;
            if (byteCount > 3) {
               buildNumber = (scardAttributeBytes[2] << 8) + scardAttributeBytes[3];
            }
            cout << "  SCARD_ATTR_VENDOR_IFD_VERSION:" << endl;
            cout << "    major: " << majorVersion << endl;
            cout << "    minor: " << minorVersion << endl;
            cout << "    build: " << buildNumber << endl;
         }
         break;

      default:
         scardAttributeBytes.push_back(0); // null-terminate this string for display.
         string stringFromBytes = (char *) &scardAttributeBytes[0];
         cout << " " << stringFromBytes << endl;
      }
   }

   // create a byte vector for the upcoming SCardTransmit() method.
   // these particular bytes should function with this type of contact chip:
   //
   //      MPCOS-EMV 16k
   //      GEMPLUS
   //      Datacard part number 806062-002
   //
   byte bytes[] = {0x00, 0xA4, 0x00, 0x00};
   vector <byte> sendBytes(bytes, bytes + ARRAYSIZE(bytes));

   cout << "sending bytes (hex): ";
   for (size_t index = 0; index < sendBytes.size(); index++) {
      cout << setw(2) << setfill('0') << uppercase << hex << int(sendBytes[index]) << " ";
   }
   cout << dec << endl;

   // for those bytes, we should receive 0x61, 0x12
   vector <byte> receivedBytes;

   if (verbose) {util::EchoCurrentTime("calling SCartTransmit()");}

   scardResult = scard.SCardTransmit(sendBytes, receivedBytes);
   cout << SCardResultMessage(scardResult, "SCardTransmit()") << endl;
   if (SCARD_S_SUCCESS == scardResult && receivedBytes.size()) {
      cout << "  " << receivedBytes.size() << " bytes received (hex): ";
      for (size_t index = 0; index < receivedBytes.size(); index++) {
         cout << setw(2) << setfill('0') << uppercase << hex << int(receivedBytes[index]) << " ";
      }
      cout << dec << endl;
   }

   receivedBytes.clear();
   sendBytes.clear();

   // send the bytes 0x00 0xC0 0x00 0x00 0x12:
   sendBytes.push_back(0x00);
   sendBytes.push_back(0xC0);
   sendBytes.push_back(0x00);
   sendBytes.push_back(0x00);
   sendBytes.push_back(0x12);

   cout << "sending bytes (hex): ";
   for (size_t index = 0; index < sendBytes.size(); index++) {
      cout << setw(2) << setfill('0') << uppercase << hex << int(sendBytes[index]) << " ";
   }
   cout << dec << endl;

   // for those bytes, we should receive
   //  0x85, 0x10, 0x80, 0x01, 0x3F, 0x00, 0x38, 0x00, 0x00, 0x00,
   //  0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x6B, 0x90, 0x00

   if (verbose) {util::EchoCurrentTime("calling SCartTransmit()");}

   scardResult = scard.SCardTransmit(sendBytes, receivedBytes);
   cout << SCardResultMessage(scardResult, "SCardTransmit()") << endl;
   if (SCARD_S_SUCCESS == scardResult && receivedBytes.size()) {
      cout << "  " << receivedBytes.size() << " bytes received (hex): ";
      for (size_t index = 0; index < receivedBytes.size(); index++) {
         cout << setw(2) << setfill('0') << uppercase << hex << int(receivedBytes[index]) << " ";
      }
      cout << dec << endl;
   }

   if (verbose) {util::EchoCurrentTime("PersonalizeChip() return");}
}

////////////////////////////////////////////////////////////////////////////////
// PersonalizeSmartcard()
// smartcard personalization using Win32 / WinSCard.H styled wrapper class.
////////////////////////////////////////////////////////////////////////////////
void PersonalizeSmartcard(CComPtr <IBidiSpl>& bidiSpl, const bool verbose)
{
   // create an instance of the wrapper for the Win32 SCard...() functions:
   dxp01sdk::SCard scard(bidiSpl);

   // try a 'contact' chip:
   DWORD protocol(SCARD_PROTOCOL_UNDEFINED);

   if (verbose) {util::EchoCurrentTime("calling SCardConnect(contact)");}

   long scardResult = scard.SCardConnect(dxp01sdk::SCard::contact, protocol);
   cout << SCardResultMessage(scardResult, "SCardConnect(contact)") << endl;
   if (SCARD_S_SUCCESS == scardResult)
   {
      cout << "SCardConnect(contact) protocol[s]: ";
      std::vector <wstring> protocols = dxp01sdk::SCard::StringsFromProtocol(protocol);
      for (size_t index = 0; index < protocols.size(); index++) {
         wcout << protocols[index] << " ";
      } cout << endl;

      PersonalizeChip(scard, verbose);

      if (verbose) {util::EchoCurrentTime("calling SCardDisConnect(contact)");}

      scardResult = scard.SCardDisConnect(SCARD_LEAVE_CARD);
      cout << SCardResultMessage(scardResult, "SCardDisConnect()") << endl << endl;
   }

   // try a 'contactless' chip:
   protocol = SCARD_PROTOCOL_UNDEFINED;

   if (verbose) {util::EchoCurrentTime("calling SCardConnect(contactless)");}

   scardResult = scard.SCardConnect(dxp01sdk::SCard::contactless, protocol);
   cout << SCardResultMessage(scardResult, "SCardConnect(contactless)") << endl;
   if (SCARD_S_SUCCESS == scardResult)
   {
      cout << "SCardConnect(contactless) protocol[s]: ";
      std::vector <wstring> protocols = dxp01sdk::SCard::StringsFromProtocol(protocol);
      for (size_t index = 0; index < protocols.size(); index++) {
         wcout << protocols[index] << " ";
      } cout << endl;

      PersonalizeChip(scard, verbose);

      if (verbose) {util::EchoCurrentTime("calling SCardDisConnect(contactless)");}

      scardResult = scard.SCardDisConnect(SCARD_LEAVE_CARD);
      cout << SCardResultMessage(scardResult, "SCardDisConnect()") << endl << endl;
   }
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   stringstream exceptionText;
   CString printerName;
   int c(0);
   bool print(false);
   bool pollForJobCompletion(false);
   wstring inputHopper(L"1");
   bool verbose(false);
   bool magstripe(false);

   while ((c = getopt(argc, argv, _T("n:pci:mv"))) != EOF) {
      switch (c) {
      case L'n': printerName = optarg; break;
      case L'i': inputHopper = optarg; break;
      case L'p': print = true; break;
      case L'c': pollForJobCompletion = true; break;
      case L'm': magstripe = true; break;
      case L'v': verbose = true; break;
      default: usage(argv[0]);
      }
   }

   if (printerName.IsEmpty()) {
      usage(argv[0]);
   }

   CComPtr <IBidiSpl> bidiSpl;
   try {
      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // create an IBidiSpl COM smartpointer instance, and bind it to the
      // printer name. We use it for all subsequent call using the IBidi interface.
      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice():" << " 0x" << hex << hr << " " << CT2A(printerName) << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // optionally display sdk version
      CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      // see if the printer is in the Printer_Ready state:
      CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);
      if ("Ready" != printerOptionsValues._printerStatus) {
         exceptionText << "printer " << CT2A(printerName) << " is not ready.";
         throw runtime_error(exceptionText.str());
      }

      // this application uses the single-wire method for communicating with
      // a smartcard reader in the printer. If the printer is not configured with
      // that option, we're done.
      if ("Single wire" != printerOptionsValues._optionSmartcard) {
         exceptionText << "printer '" << CT2A(printerName) << "' does not have single-wire smartcard option. Exiting." << endl;;
         throw runtime_error(exceptionText.str());
      }

      StartJob(bidiSpl, verbose, inputHopper);

      // park the card in the smartcard reader:
      util::PrinterStatusValues printerStatusValues = ParkCard(bidiSpl, verbose);

      // the card has been parked in the smartcard reader.
      bool personalizedOK(false);
      try {
         PersonalizeSmartcard(bidiSpl, verbose);
         personalizedOK = true;
      }
      catch (runtime_error) {
         // already logged
      }

      if (!personalizedOK) {
         cout << "smartcard personalization failed." << endl;
         CancelCard(bidiSpl, printerStatusValues, verbose);
      }
      else {
         cout << "smartcard personalization succeeded." << endl;

         ResumeCard(printerStatusValues._printerJobID, bidiSpl, verbose);

         if (magstripe) {
            EncodeMagstripe(bidiSpl, verbose);
         }

         if (print) {

            // do some simple text and graphics printing.

            const bool doTwoSided = 
               printerOptionsValues._optionDuplex == "Auto" ||
               printerOptionsValues._optionDuplex == "Manual";

            if (verbose) {util::EchoCurrentTime("calling PrintText()");}

            // Important: util::PrintTextmethod also waits until driver gets all the print data.
            util::PrintText(printerName, doTwoSided);

            if (verbose) {util::EchoCurrentTime("PrintText() returned");}
         }
      }

      EndJob(bidiSpl, verbose);

      if (pollForJobCompletion) {
         PollForJobCompletion(bidiSpl, printerStatusValues._printerJobID, verbose);
      }
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}
