////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps card printer sdk c++ 'status' sample.
////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <iostream>
#include <sstream>
#include <bidispl.h>
#include "XGetopt.h"
#include "util.h"
#include "DXP01SDK.H"

using namespace std;

struct CommandLineOptions
{
   CString printerName;
   bool    jobStatus;
   CommandLineOptions() : jobStatus(false) {}
};

CString GetShortExeName()
{
   CString exeName;
   ::GetModuleFileName(NULL, exeName.GetBuffer(MAX_PATH), MAX_PATH);
   CPath shortExeName(exeName);
   shortExeName.StripPath();
   return shortExeName;
}

void usage()
{
   cout << endl;
   cout << CT2A(GetShortExeName()) << " demonstrates using interactive mode to retrieve" << endl;
   cout << "printer options, printer messages, and job status." << endl << endl;
   cout << CT2A(GetShortExeName()) << " -n <printername>" << endl << endl;
   cout << "options:" << endl;
   cout << "  " << "-n <printername>. Required. Try -n \"XPS Card Printer\"." << endl;
   cout << "  " << "-j Optional. Submit a card job and display job status." << endl << endl;
   cout << CT2A(GetShortExeName()) << " -n \"XPS Card Printer\"" << endl;
   cout << "Retrieves printer options, starts a job, retrieves printer messages," << endl;
   cout << "ends the job, and retrieves job status." << endl;
   cout << endl;
   ::exit(-1);
}

void CancelCard(
   CComPtr <IBidiSpl>&              bidiSpl,
   const util::PrinterStatusValues& printerStatusValues)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // create the XML and convert to bytes for the upcoming SetInputData() call.
   // NOTE: It is IMPORTANT to set the errorcode member; it clears that
   // particular error in the printer.
   CString actionXML = util::FormatPrinterActionXML(
      dxp01sdk::Cancel,
      printerStatusValues._printerJobID,
      printerStatusValues._errorCode);

   const int xmlBytesLength = actionXML.GetLength() * sizeof WCHAR;

   // allocate IBidiSpl compatible memory:
   CComAllocator XMLBytesAllocator;
   PBYTE xmlBytes = (PBYTE) XMLBytesAllocator.Allocate(xmlBytesLength);
   DBG_UNREFERENCED_LOCAL_VARIABLE(XMLBytesAllocator); // suppress annoying compiler warning

   // copy the XML string...without the trailing NULL:
   ::memcpy_s(xmlBytes, xmlBytesLength, actionXML.GetBuffer(), xmlBytesLength);

   hr = bidiRequest->SetSchema(dxp01sdk::PRINTER_ACTION);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::PRINTER_ACTION): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetInputData(BIDI_BLOB, xmlBytes, xmlBytesLength);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest->SetInputData(): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) cancel xml: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }
}

void EndJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "CoCreateInstance(CLSID_BidiRequest): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::ENDJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::ENDJOB): " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);         
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) endjob: " << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   // without any printing in this sample app, EndJob() finishes very quickly.
   // We might fail on getting the printer status.
}

void DisplayPrinterStatus(const util::PrinterStatusValues& statusVals)
{
   cout << endl << "Status:" << endl;
   cout << "  ClientID:        " << CW2A(statusVals._clientID) << endl;
   cout << "  ErrorCode:       " << statusVals._errorCode << endl;
   cout << "  ErrorSeverity:   " << statusVals._errorSeverity << endl;
   cout << "  ErrorString:     " << CW2A(statusVals._errorString) << endl;
   cout << "  PrinterData:     " << CW2A(statusVals._printerData) << endl;
   cout << "  PrinterJobID:    " << statusVals._printerJobID << endl;
   cout << "  WindowsJobID:    " << statusVals._windowsJobID << endl;
}

void DisplayPrinterCardCount(const util::PrinterCardCountValues& countVals)
{
   cout << endl << "Counts:"  << endl;
   cout << "  CardsPickedSinceCleaningCard: " <<      countVals._cardsPickedSinceCleaningCard << endl;
   cout << "  CleaningCardsRun:             " <<      countVals._cleaningCardsRun             << endl;
   cout << "  CurrentCompleted:             " <<      countVals._currentCompleted             << endl;
   cout << "  CurrentLost:                  " <<      countVals._currentLost                  << endl;
   cout << "  CurrentPicked:                " <<      countVals._currentPicked                << endl;
   cout << "  CurrentPickedExceptionSlot:   " <<      countVals._currentPickedExceptionSlot   << endl;
   cout << "  CurrentPickedInputHopper1:    " <<      countVals._currentPickedInputHopper1    << endl;
   cout << "  CurrentPickedInputHopper2:    " <<      countVals._currentPickedInputHopper2    << endl;
   cout << "  CurrentPickedInputHopper3:    " <<      countVals._currentPickedInputHopper3    << endl;
   cout << "  CurrentPickedInputHopper4:    " <<      countVals._currentPickedInputHopper4    << endl;
   cout << "  CurrentPickedInputHopper5:    " <<      countVals._currentPickedInputHopper5    << endl;
   cout << "  CurrentPickedInputHopper6:    " <<      countVals._currentPickedInputHopper6    << endl;
   cout << "  CurrentRejected:              " <<      countVals._currentRejected              << endl;
   cout << "  PrinterStatus:                " << CW2A(countVals._printerStatus)               << endl;
   cout << "  TotalCompleted:               " <<      countVals._totalCompleted               << endl;
   cout << "  TotalLost:                    " <<      countVals._totalLost                    << endl;
   cout << "  TotalPicked:                  " <<      countVals._totalPicked                  << endl;
   cout << "  TotalPickedExceptionSlot:     " <<      countVals._totalPickedExceptionSlot     << endl;
   cout << "  TotalPickedInputHopper1:      " <<      countVals._totalPickedInputHopper1      << endl;
   cout << "  TotalPickedInputHopper2:      " <<      countVals._totalPickedInputHopper2      << endl;
   cout << "  TotalPickedInputHopper3:      " <<      countVals._totalPickedInputHopper3      << endl;
   cout << "  TotalPickedInputHopper4:      " <<      countVals._totalPickedInputHopper4      << endl;
   cout << "  TotalPickedInputHopper5:      " <<      countVals._totalPickedInputHopper5      << endl;
   cout << "  TotalPickedInputHopper6:      " <<      countVals._totalPickedInputHopper6      << endl;
   cout << "  TotalRejected:                " <<      countVals._totalRejected                << endl;
}

void DisplayPrinterOptions(const util::PrinterOptionsValues& optionVals)
{
   cout << endl << "Options:" << endl;
   cout << "  ColorPrintResolution:       " << CW2A(optionVals._colorPrintResolution) << endl;
   cout << "  ConnectionPortType:         " << CW2A(optionVals._connectionPortType) << endl;
   cout << "  ConnectionProtocol:         " << CW2A(optionVals._connectionProtocol) << endl;
   cout << "  EmbosserVersion:            " << CW2A(optionVals._embosserVersion) << endl;
   cout << "  LockState:                  " << CW2A(optionVals._lockState) << endl;
   cout << "  ModuleEmbosser:             " << CW2A(optionVals._moduleEmbosser) << endl;
   cout << "  MonochromePrintResolution:  " << CW2A(optionVals._monochromePrintResolution) << endl;
   cout << "  OptionDuplex:               " << CW2A(optionVals._optionDuplex) << endl;
   cout << "  OptionInputhopper:          " << CW2A(optionVals._optionInputhopper) << endl;
   cout << "  OptionLocks:                " << CW2A(optionVals._optionLocks) << endl;
   cout << "  OptionMagstripe:            " << CW2A(optionVals._optionMagstripe) << endl;
   cout << "  OptionSmartcard:            " << CW2A(optionVals._optionSmartcard) << endl;
   cout << "  PrinterAddress:             " << CW2A(optionVals._printerAddress) << endl;
   cout << "  PrinterMessageNumber:       " <<      optionVals._printerMessageNumber << endl;
   cout << "  PrinterModel:               " << CW2A(optionVals._printerModel) << endl;
   cout << "  PrinterSerialNumber:        " << CW2A(optionVals._printerSerialNumber) << endl;
   cout << "  PrinterStatus:              " << CW2A(optionVals._printerStatus) << endl;
   cout << "  PrinterVersion:             " << CW2A(optionVals._printerVersion) << endl;
   cout << "  PrintHead:                  " << CW2A(optionVals._printHead) << endl;
   cout << "  TopcoatPrintResolution:     " << CW2A(optionVals._topcoatPrintResolution) << endl;
}

void DisplaySupplies(const util::SuppliesValues& supplyvals) 
{
   cout << endl << "Supplies:" << endl;
   cout << "  IndentRibbon:            " << CW2A(supplyvals._indentRibbon) << endl;
   cout << "  IndentRibbonRemaining:   " <<      supplyvals._indentRibbonRemaining << "%" << endl;
   cout << "  PrinterStatus:           " << CW2A(supplyvals._printerStatus) << endl;
   cout << "  PrintRibbonLotCode:      " << CW2A(supplyvals._printRibbonLotCode)  << endl;
   cout << "  PrintRibbonPartNumber:   " <<      supplyvals._printRibbonPartNumber << endl;
   cout << "  PrintRibbonSerialNumber: " << CW2A(supplyvals._printRibbonSerialNumber)  << endl;
   cout << "  PrintRibbonType:         " << CW2A(supplyvals._printRibbonType) << endl;
   cout << "  RibbonRemaining:         " <<      supplyvals._ribbonRemaining << "%" << endl;
   cout << "  TopperRibbonRemaining:   " <<      supplyvals._topperRibbonRemaining << "%" << endl;
   cout << "  TopperRibbonType:        " << CW2A(supplyvals._topperRibbonType) << endl;
}

void DisplayJobStatus(const util::JobStatusValues& jobsStatusValues)
{
   cout << endl << "Job Status:" << endl;
   cout << "  ClientID:        " << CW2A(jobsStatusValues._clientID) << endl;
   cout << "  JobRestartCount: " << jobsStatusValues._jobRestartCount << endl;
   cout << "  JobState:        " << CW2A(jobsStatusValues._jobState) << endl;
   cout << "  PrinterJobID:    " << jobsStatusValues._printerJobID << endl;
   cout << "  WindowsJobID:    " << jobsStatusValues._windowsJobID << endl;
}

CString StartJob(CComPtr <IBidiSpl>& bidiSpl)
{
   stringstream exceptionText;

   CComPtr <IBidiRequest> bidiRequest;
   HRESULT hr = bidiRequest.CoCreateInstance(CLSID_BidiRequest);
   if (FAILED(hr)) {
      exceptionText << "bidiRequest.CoCreateInstance(CLSID_BidiRequest):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiRequest->SetSchema(dxp01sdk::STARTJOB);
   if (FAILED(hr)) {
      exceptionText << "SetSchema(dxp01sdk::STARTJOB):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   hr = bidiSpl->SendRecv(BIDI_ACTION_SET, bidiRequest);
   if (FAILED(hr)) {
      exceptionText << "SendRecv(BIDI_ACTION_SET) startjob:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   CString printerStatusXML = util::GetPrinterStatusXML(bidiRequest);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   if (0 != printerStatusValues._errorCode) {
      return printerStatusXML;
   }

   HRESULT startJobResultCode(S_OK);
   hr = bidiRequest->GetResult(&startJobResultCode);
   if (FAILED(hr)) {
      exceptionText << "GetResult(startjob):" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
      throw runtime_error(exceptionText.str());
   }

   if (FAILED(startJobResultCode)) {
      exceptionText << "startJobResultCode:" << " 0x" << hex << startJobResultCode << " " << CT2A(util::Win32ErrorString(startJobResultCode));
      throw runtime_error(exceptionText.str());
   }

   return printerStatusXML;
}

void ValidateCommandLineOptions(CommandLineOptions& commandLineOptions)
{
   if (commandLineOptions.printerName.IsEmpty()) {usage();}
}

CommandLineOptions GetCommandlineOptions(const int argc, _TCHAR* argv[])
{
   CommandLineOptions commandLineOptions;
   int c(0);
   while ((c = getopt(argc, argv, _T("n:j"))) != EOF) {
      switch (c) {
      case L'n': commandLineOptions.printerName = optarg; break;
      case L'j': commandLineOptions.jobStatus = true; break;
      default: usage();
      }
   }
   return commandLineOptions;
}

void CardJobWithStatus(CComPtr <IBidiSpl>& bidiSpl)
{
   CString printerStatusXML = StartJob(bidiSpl);
   util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXML);
   DisplayPrinterStatus(printerStatusValues);

   if (0 != printerStatusValues._errorCode) {
      CancelCard(bidiSpl, printerStatusValues);
      cerr << "StartJob() failed:" << " 0x" << printerStatusValues._errorCode << " " << CT2A(printerStatusValues._errorString) << endl;
      throw printerStatusValues;
   }

   const int printerJobID = printerStatusValues._printerJobID;

   EndJob(bidiSpl);

   // poll for job status and display values:
   util::JobStatusValues jobStatusValues;
   do {
      CString jobStatusXml = util::GetJobStatusXML(bidiSpl, printerJobID);
      jobStatusValues = util::ParseJobStatusXML(jobStatusXml);
      DisplayJobStatus(jobStatusValues);
      ::Sleep(1000);
   } while (jobStatusValues._jobState == dxp01sdk::JOB_ACTIVE);

   cout << endl;
}

////////////////////////////////////////////////////////////////////////////////
// main()
////////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[])
{
   stringstream exceptionText;

   CommandLineOptions commandLineOptions = GetCommandlineOptions(argc, argv);
   ValidateCommandLineOptions(commandLineOptions);

   CString& printerName = commandLineOptions.printerName;  // renamed for convenience

   CComPtr <IBidiSpl> bidiSpl;
   try
   {
      HRESULT hr = ::CoInitialize(NULL);
      if (FAILED(hr)) {
         exceptionText << "::CoInitialize:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      // create an IBidiSpl COM smartpointer instance, and bind it to the
      // printer name. We use it for all subsequent call using the IBidi interface.
      hr = bidiSpl.CoCreateInstance(CLSID_BidiSpl);
      if (FAILED(hr)) {
         exceptionText << "bidiSpl.CoCreateInstance:" << " 0x" << hex << hr << " " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      hr = bidiSpl->BindDevice(printerName, BIDI_ACCESS_USER);
      if (FAILED(hr)) {
         exceptionText << "BindDevice() error " << " 0x" << hex << hr << " '" << CT2A(printerName) << "' " << CT2A(util::Win32ErrorString(hr));
         throw runtime_error(exceptionText.str());
      }

      CString sdkVersionXml = util::GetSDKVersionXML(bidiSpl);
      CString sdkVersionString = util::ParseSDKVersionXML(sdkVersionXml);
      cout << endl << "SDK Version: " << CW2A(sdkVersionString) << endl << endl;

      const CString printerStatusXml = util::GetPrinterStatusXML(bidiSpl);
      const util::PrinterStatusValues printerStatusValues = util::ParsePrinterStatusXML(printerStatusXml);
      DisplayPrinterStatus(printerStatusValues);

      const CString printerOptionsXML = util::GetPrinterOptionsXML(bidiSpl);
      const util::PrinterOptionsValues printerOptionsValues = util::ParsePrinterOptionsXML(printerOptionsXML);
      DisplayPrinterOptions(printerOptionsValues);

      const CString printerCardCountsXML = util::GetPrinterCardCountsXML(bidiSpl);
      const util::PrinterCardCountValues printerCardCountValues = util::ParsePrinterCardCountsXML(printerCardCountsXML);
      DisplayPrinterCardCount(printerCardCountValues);

      const CString printerSuppliesXML = util::GetSuppliesXML(bidiSpl);
      util::SuppliesValues suppliesValues = util::ParseSuppliesXML(printerSuppliesXML);
      DisplaySupplies(suppliesValues);

      if (commandLineOptions.jobStatus) {
         CardJobWithStatus(bidiSpl);
      }
   }
   catch (util::PrinterStatusValues) {
      // already reported
   }
   catch (runtime_error& e) {
      cerr << e.what() << endl;
   }

   if (bidiSpl) {
      bidiSpl->UnbindDevice();
      bidiSpl = NULL;
   }

   ::CoUninitialize();

   return 0;
}
