﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
//  xps card printer SDK: c sharp 'card completion' sample
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing;
using System.Drawing.Printing;
using System.IO;
using System.Threading;

using dxp01sdk;

class CommandLineOptions
{
    public string printerName = string.Empty;
}

class CardCompletionDocument : PrintDocument
{
    public CardCompletionDocument(CommandLineOptions commandLineOptions)
    {
        DocumentName = Path.GetFileName(Environment.GetCommandLineArgs()[0]) + " printing";
        PrinterSettings.PrinterName = commandLineOptions.printerName;
        PrintController = new StandardPrintController(); // suppress annoying progress popups
        PrintPage += new PrintPageEventHandler(OnPrintPage);
    }

    private void OnPrintPage(object sender, PrintPageEventArgs pageEventArgs)
    {
        using (Font font = new Font("Arial", 8))
        using (SolidBrush brush = new SolidBrush(Color.Black))
        using (Pen pen = new Pen(Color.DarkBlue))
        {
            pageEventArgs.Graphics.DrawString("Sample Text", font, brush, 50, 50);
            pageEventArgs.Graphics.DrawRectangle(pen, 50, 70, 20, 25);
        }
    }
}

class card_completion
{
    private static void Usage()
    {
        string thisExeName = Path.GetFileName(Environment.GetCommandLineArgs()[0]);
        Console.WriteLine(thisExeName + " demonstrates the card completion for job");
        Console.WriteLine("printed using Windows printing methods.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n <printername>");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine();
        Environment.Exit(-1);
    }

    private static CommandLineOptions GetCommandlineOptions(string[] args)
    {
        if (args.Length == 0) Usage();

        CommandLineOptions commandLineOptions = new CommandLineOptions();
        CommandLine.Utility.Arguments arguments = new CommandLine.Utility.Arguments(args);

        if (string.IsNullOrEmpty(arguments["n"])) { Usage(); }
        // we might have a -n with no printer name:
        bool boolVal = false;
        if (Boolean.TryParse(arguments["n"], out boolVal)) { Usage(); }
        commandLineOptions.printerName = arguments["n"];

        return commandLineOptions;
    }

    private static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
    {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    static void PollForJobCompletion(BidiSplWrap bidiSpl, int printerJobID)
    {
        for (; true; )
        {
            string temp = strings.JOB_STATUS_XML;
            string jobInfoXML = string.Format(temp, printerJobID);
            string jobStatusXML = bidiSpl.GetPrinterData(strings.JOB_STATUS, jobInfoXML);
            JobStatusValues jobStatusValues = Util.ParseJobStatusXML(jobStatusXML);
            Console.WriteLine("Printer job id: " + printerJobID + " " + jobStatusValues._jobState);
            if (strings.JOB_ACTIVE != jobStatusValues._jobState)
            {
                return;
            }
            Thread.Sleep(1000);
        }
    }

    // repeatedly get the printer status until with printer status contains
    // a valid printer jobID.
    static int WaitForPrinterJobID(BidiSplWrap bidiSpl, string printerName)
    {
        const int sleepMilliSeconds = 500;

        var printerStatusXML = bidiSpl.GetPrinterData(strings.PRINTER_MESSAGES);
        var printerStatus = Util.ParsePrinterStatusXML(printerStatusXML);

        do
        {
            if (0 != printerStatus._printerJobID)
            {
                return printerStatus._printerJobID;
            }

            Console.WriteLine(string.Format("waiting {0} millseconds for printer jobID", sleepMilliSeconds));
            Thread.Sleep(500);

            printerStatusXML = bidiSpl.GetPrinterData(strings.PRINTER_MESSAGES);
            printerStatus = Util.ParsePrinterStatusXML(printerStatusXML);

        } while (0 == printerStatus._errorCode);

        Console.WriteLine("printer error:");
        Console.WriteLine(printerStatusXML);

        return printerStatus._printerJobID;
    }

    public static void Main(string[] args)
    {
        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = new BidiSplWrap();
        bidiSpl.BindDevice(commandLineOptions.printerName);

        // optional:
        string sdkVersion = bidiSpl.GetPrinterData(strings.SDK_VERSION);
        Console.WriteLine("\nSDK Version: " + Util.ParseSDKVersionXML(sdkVersion) + "\n");

        // optional: is printer is ready for an Ibidi StartJob()?
        string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);
        PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);
        if ("Ready" != printerOptionsValues._printerStatus)
        {
            Console.WriteLine(commandLineOptions.printerName + " is not ready.");
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        CardCompletionDocument printDocument = new CardCompletionDocument(commandLineOptions);
        printDocument.Print();

        // get the printer job id:
        int printerJobID = WaitForPrinterJobID(bidiSpl, commandLineOptions.printerName);

        // poll for job completion:
        PollForJobCompletion(bidiSpl, printerJobID);

        bidiSpl.UnbindDevice();
    }
}
