﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps card printer SDK: csharp 'emboss_indent' sample
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing;
using System.Drawing.Printing;
using System.IO;
using System.Threading;
using dxp01sdk;

class CommandLineOptions
{
    public string printerName = string.Empty;
    public string inputHopper = "1";
    public bool pollForJobCompletion = false;
}

class EmbossIndenntDocument : PrintDocument
{
    public EmbossIndenntDocument(CommandLineOptions commandLineOptions)
    {
        DocumentName = "emboss indent document";
        PrinterSettings.PrinterName = commandLineOptions.printerName;
        PrintController = new StandardPrintController(); // suppress annoying progress popups
        PrintPage += new PrintPageEventHandler(OnPrintPage);
    }

    void OnPrintPage(object sender, PrintPageEventArgs pageEventArgs)
    {
        string text = string.Empty;

        // emit all emboss + indent data driver escape strings. the font and 
        // brush are not significant; just needed for the DrawString() call.
        using (Font font = new Font("Arial", 8))
        using (SolidBrush brush = new SolidBrush(Color.Black))
        {
            text = "~EM%1;301;860;Font 11111";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%2;1600;860;222222";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%3;301;1460;333333";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%4;301;1180;444444";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%5;301;690;555555";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%6;1600;690;666666";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%7;301;650;777777";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%8;301;1000;888888";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%9;301;1050;999999";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);

            text = "~EM%10;1600;1050;10 10 10";
            pageEventArgs.Graphics.DrawString(text, font, brush, 50, 50);
        }
    }
}

class emboss_indent
{
    static void Usage()
    {
        string thisExeName = Path.GetFileName(Environment.GetCommandLineArgs()[0]);
        Console.WriteLine(thisExeName + " demonstrates the emboss and indent escapes, input hopper");
        Console.WriteLine("selection, and the checking of print ribbon and emboss supplies.");
        Console.WriteLine();
        Console.WriteLine("Emboss and indent data is hard coded.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n <printername> [-i <input hopper>] [-c]");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine("  -i <input hopper>. Optional. Defaults to \"1\".");
        Console.WriteLine("  -c Poll for job status. Optional. Defaults to false.");
        Console.WriteLine();
        Environment.Exit(-1);
    }

    static CommandLineOptions GetCommandlineOptions(string[] args)
    {
        if (args.Length == 0) Usage();

        CommandLineOptions commandLineOptions = new CommandLineOptions();
        CommandLine.Utility.Arguments arguments = new CommandLine.Utility.Arguments(args);

        if (string.IsNullOrEmpty(arguments["n"])) { Usage(); }
        // we might have a -n with no printer name:
        bool boolVal = false;
        if (Boolean.TryParse(arguments["n"], out boolVal)) { Usage(); }
        commandLineOptions.printerName = arguments["n"];

        if (!string.IsNullOrEmpty(arguments["i"]))
        {
            // we might have a -i with no input hopper:
            if (Boolean.TryParse(arguments["i"], out boolVal)) { Usage(); }
            commandLineOptions.inputHopper = arguments["i"];
        }

        commandLineOptions.pollForJobCompletion = !string.IsNullOrEmpty(arguments["c"]);
        return commandLineOptions;
    }

    static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
    {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    static void PollForJobCompletion(BidiSplWrap bidiSpl, int printerJobID)
    {
        for (; true; )
        {
            string temp = strings.JOB_STATUS_XML;
            string jobInfoXML = string.Format(temp, printerJobID);
            string jobStatusXML = bidiSpl.GetPrinterData(strings.JOB_STATUS, jobInfoXML);
            JobStatusValues jobStatusValues = Util.ParseJobStatusXML(jobStatusXML);
            Console.WriteLine("Printer job id: " + printerJobID + " " + jobStatusValues._jobState);
            if (strings.JOB_ACTIVE != jobStatusValues._jobState)
            {
                return;
            }
            Thread.Sleep(1000);
        }
    }

    /// <summary>
    /// for this sample, pass the hopper name from the command line into the
    /// IBidi StartJob() call.
    /// </summary>
    /// <param name="inputHopper">
    /// a string like "1", "Exception", or empty string
    /// </param>
    /// <returns>
    /// the printer Job ID, an integer
    /// </returns>
    static int StartJob(BidiSplWrap bidiSpl, string inputHopper)
    {
        string startJobXml = string.Format(
            strings.STARTJOB_XML,
            inputHopper,
            0,      // zero means "do not check print ribbon"
            0);     // zero means "do not check emboss supplies"

        string printerStatusXML = bidiSpl.SetPrinterData(strings.STARTJOB, startJobXml);
        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            // the StartJob failed; send a Cancel(); do NOT send an EndJob():
            CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode);
            Console.WriteLine("Canceling the job. Printer Error: " + printerStatusValues._errorString);
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }
        return printerStatusValues._printerJobID;
    }

    static void Main(string[] args)
    {
        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = new BidiSplWrap();
        bidiSpl.BindDevice(commandLineOptions.printerName);

        // optional:
        string sdkVersion = bidiSpl.GetPrinterData(strings.SDK_VERSION);
        Console.WriteLine("\nSDK Version: " + Util.ParseSDKVersionXML(sdkVersion) + "\n");

        // optional: is printer is ready for an Ibidi StartJob()?
        string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);
        PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);
        if ("Ready" != printerOptionsValues._printerStatus)
        {
            Console.WriteLine(commandLineOptions.printerName + " is not ready.");
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        int printerJobID = StartJob(bidiSpl, commandLineOptions.inputHopper);

        var printDocument = new EmbossIndenntDocument(commandLineOptions);
        printDocument.Print();
        Util.WaitForWindowsJobID(bidiSpl, commandLineOptions.printerName);

        bidiSpl.SetPrinterData(strings.ENDJOB);

        if (commandLineOptions.pollForJobCompletion)
        {
            PollForJobCompletion(bidiSpl, printerJobID);
        }

        bidiSpl.UnbindDevice();
    }
}
