﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps printer sdk magstripe sample.
////////////////////////////////////////////////////////////////////////////////
using System;
using System.IO;
using System.Threading;
using dxp01sdk;

public class CommandLineOptions
{
    public string printerName;
    public bool readMagstripe;
    public bool print;
    public bool pollForJobCompletion;
    public bool jisRequest;
    public bool verbose;
};

class magstripe
{
    private static void usage()
    {
        Console.WriteLine("Magstripe.exe demonstrates interactive mode magnetic stripe encoding with");
        Console.WriteLine("  options to read the magnetic stripe data, print, and poll for job");
        Console.WriteLine("  completion status.");
        Console.WriteLine();
        Console.WriteLine("> Uses hard-coded data for magnetic stripe and printing.");
        Console.WriteLine();
        Console.WriteLine("magstripe.exe -n <printername> [-r] [-p] [-c] [-j] [-v]");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine("  -r Reads back the encoded magnetic stripe data.");
        Console.WriteLine("  -p Print simple black text on front side of card. ");
        Console.WriteLine("  -c Poll for job status. ");
        Console.WriteLine("  -j JIS magnetic. ");
        Console.WriteLine("  -v Display detailed information with timings.");
        Console.WriteLine();
        Console.WriteLine("magstripe.exe -n \"XPS Card Printer\"");
        Console.WriteLine("  Encodes data on all three tracks of an ISO 3-track magnetic stripe on the");
        Console.WriteLine("  backside of a card.");
        Console.WriteLine();
        Console.WriteLine("magstripe -n \"XPS Card Printer\" -r -p -c");
        Console.WriteLine("  Encodes data on all three tracks of an ISO 3-track magnetic stripe on the");
        Console.WriteLine("  backside of a card, reads the data from the magnetic stripe and displays");
        Console.WriteLine("  it, prints black text on the front side of the card, and polls and");
        Console.WriteLine("  displays job status.");
        Environment.Exit(-1);
    }

    static CommandLineOptions GetCommandlineOptions(string[] args)
    {
        if (args.Length == 0) usage();

        CommandLineOptions commandLineOptions = new CommandLineOptions();
        CommandLine.Utility.Arguments arguments = new CommandLine.Utility.Arguments(args);

        if (!string.IsNullOrEmpty(arguments["h"])) usage();

        if (string.IsNullOrEmpty(arguments["n"]))
        {
            Console.WriteLine("printer name is required");
            Environment.Exit(-1);
        }
        else
        {
            // we might have a -n with no printer name:
            bool boolVal = false;
            if (Boolean.TryParse(arguments["n"], out boolVal))
            {
                Console.WriteLine("printer name is required");
                Environment.Exit(-1);
            }
            commandLineOptions.printerName = arguments["n"];
        }

        commandLineOptions.readMagstripe = !string.IsNullOrEmpty(arguments["r"]);
        commandLineOptions.print = !string.IsNullOrEmpty(arguments["p"]);
        commandLineOptions.pollForJobCompletion = !string.IsNullOrEmpty(arguments["c"]);
        commandLineOptions.jisRequest = !string.IsNullOrEmpty(arguments["j"]);
        commandLineOptions.verbose = !string.IsNullOrEmpty(arguments["v"]);

        return commandLineOptions;
    }

    static void ResumeJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
    {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Resume, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " CancelJob() start"); }

        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " CancelJob() return"); }
    }

    static void PollForJobCompletion(BidiSplWrap bidiSpl, int printerJobID, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PollForJobCompletion() start"); }

        for (; true; )
        {

            string printerStatusXML = bidiSpl.GetPrinterData(strings.PRINTER_MESSAGES);
            PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
            if (0 != printerStatusValues._errorCode)
            {
                Console.WriteLine(printerStatusValues._errorString + " severity: " + printerStatusValues._errorSeverity.ToString());
                Console.WriteLine("canceling printer jobID " + printerStatusValues._printerJobID.ToString());
                CancelJob(
                    bidiSpl,
                    printerStatusValues._printerJobID,
                    printerStatusValues._errorCode,
                    verbose);
                return;
            }

            string xmlFormat = strings.JOB_STATUS_XML;
            string jobInfoXML = string.Format(xmlFormat, printerJobID);
            string jobStatusXML = bidiSpl.GetPrinterData(strings.JOB_STATUS, jobInfoXML);
            JobStatusValues jobStatusValues = Util.ParseJobStatusXML(jobStatusXML);

            Console.WriteLine("Printer job id: " + printerJobID + " " + jobStatusValues._jobState);

            if (strings.JOB_ACTIVE != jobStatusValues._jobState)
            {
                return;
            }
            Thread.Sleep(1000);
        }
    }

    public static void ReadMagstripe(BidiSplWrap bidiSpl, bool jisRequest, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " ReadMagstripe() caling bidi GetPrinterData()"); }

        // replace schema string MAGSTRIPE_READ to MAGSTRIPE_READ_FRONT for front side read
        string printerStatusXML = bidiSpl.GetPrinterData(strings.MAGSTRIPE_READ);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " GetPrinterData() return"); }

        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 == printerStatusValues._errorCode)
        {
            Console.WriteLine("Magstripe data read successfully; printer job id: " + printerStatusValues._printerJobID);

            string track1 = "";
            string track2 = "";
            string track3 = "";

            Util.ParseMagstripeStrings(printerStatusXML, ref track1, ref track2, ref track3, jisRequest);

            if (track1.Length != 0)
            {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track1);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track1 Base64 decoded: " + str);
            }

            if (track2.Length != 0)
            {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track2);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track2 Base64 decoded: " + str);
            }

            if (track3.Length != 0)
            {
                // Convert the Base64 UUEncoded output.
                byte[] binaryData = System.Convert.FromBase64String(track3);
                string str = System.Text.Encoding.UTF8.GetString(binaryData);
                Console.WriteLine(" track3 Base64 decoded: " + str);
            }
        }
        else
        {
            Console.WriteLine("ReadMagstripe() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);

            CancelJob(
                bidiSpl,
                printerStatusValues._printerJobID,
                printerStatusValues._errorCode,
                verbose);

            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }
    }

    public static void EncodeMagstripe(
        BidiSplWrap bidiSpl,
        bool jisRequest,
        bool verbose)
    {
        // Hardcoded XML to encode all 3 tracks in IAT mode.
        // track 1 = "TRACK1", track 2 = "1122", track 3 = "321"
        string trackDataXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
         + "<magstripe>"
         + "<track number=\"1\"><base64Data>VFJBQ0sx</base64Data></track>"
          + "<track number=\"2\"><base64Data>MTEyMg==</base64Data></track>"
          + "<track number=\"3\"><base64Data>MzIx</base64Data></track>"
         + "</magstripe>";

        if (jisRequest)
        {
            // JIS only allows track 3 = "321"
            trackDataXML = "<?xml version=\"1.0\" encoding=\"utf-8\"?>"
          + "<magstripe>"
          + "<track number=\"1\"><base64Data></base64Data></track>"
           + "<track number=\"2\"><base64Data></base64Data></track>"
           + "<track number=\"3\"><base64Data>MzIx</base64Data></track>"
          + "</magstripe>";
        }

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " EncodeMagstripe() caling bidi SetPrinterData()"); }

        // replace schema string MAGSTRIPE_ENCODE to MAGSTRIPE_ENCODE_FRONT for front side encode
        string printerStatusXML = bidiSpl.SetPrinterData(strings.MAGSTRIPE_ENCODE, trackDataXML);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SetPrinterData() return"); }

        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("EncodeMagstripe() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);

            CancelJob(
                bidiSpl,
                printerStatusValues._printerJobID,
                printerStatusValues._errorCode,
                verbose);

            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " EncodeMagstripe() return"); }
    }

    static void StartJob(BidiSplWrap bidiSpl, ref int printerJobID, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " StartJob() begin"); }

        string startJobXml = string.Format(
            strings.STARTJOB_XML,
            "",     // use default input hopper
            0,      // zero means "do not check print ribbon"
            0);     // zero means "do not check emboss supplies"

        string printerStatusXML = bidiSpl.SetPrinterData(strings.STARTJOB, startJobXml);

        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("StartJob() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);

            CancelJob(
                bidiSpl,
                printerStatusValues._printerJobID,
                printerStatusValues._errorCode,
                verbose);

            bidiSpl.UnbindDevice();

            Environment.Exit(-1);
        }
        else
        {
            printerJobID = printerStatusValues._printerJobID;
        }

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " StartJob() return"); }
    }

    ////////////////////////////////////////////////////////////////////////////////
    // Main()
    ////////////////////////////////////////////////////////////////////////////////
    static void Main(string[] args)
    {
        int printerJobID = 0;
        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = new BidiSplWrap();
        bidiSpl.BindDevice(commandLineOptions.printerName);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling GetPrinterData(version)"); }

        string sdkVersionXml = bidiSpl.GetPrinterData(strings.SDK_VERSION);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " GetPrinterData(version) return"); }

        // Check if printer is ready to accept magstripe job
        string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);
        PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);
        if ("Ready" != printerOptionsValues._printerStatus)
        {
            Console.WriteLine(commandLineOptions.printerName + " is not ready.");
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        // Start the magstripe job
        StartJob(bidiSpl, ref printerJobID, commandLineOptions.verbose);

        // Encode the magstripe
        EncodeMagstripe(bidiSpl, commandLineOptions.jisRequest, commandLineOptions.verbose);

        // Check the command line parameters if user wants to read the magstripe
        if (commandLineOptions.readMagstripe)
        {
            // Read the magstripe
            ReadMagstripe(bidiSpl, commandLineOptions.jisRequest, commandLineOptions.verbose);
        }

        // Check if user wants print
        if (commandLineOptions.print)
        {
            if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling PrintTextAndGraphics()"); }
            Util.PrintTextAndGraphics(bidiSpl, commandLineOptions.printerName);
            if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PrintTextAndGraphics() returned"); }
        }


        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling EndJob()"); }

        // End the magstripe job
        bidiSpl.SetPrinterData(strings.ENDJOB);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " EndJob() returned"); }

        // Check if user wants to poll the job for completion status
        if (commandLineOptions.pollForJobCompletion)
        {
            PollForJobCompletion(bidiSpl, printerJobID, commandLineOptions.verbose);
        }

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling UnbindDevice()"); }

        bidiSpl.UnbindDevice();

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " UnbindDevice() returned"); }
    }
}
