﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps driver sdk csharp print sample.
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Drawing.Printing;
using dxp01sdk;
using System.IO;

namespace print
{
    class print
    {
        static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
        {
            string xmlFormat = strings.PRINTER_ACTION_XML;
            string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
            bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
        }

        /// <summary>
        /// issue an IBidi StartJob with xml markup to start an IBidi job, just to have
        /// the printer check for supplies. If the printer is out of supplies, throw an
        /// exception.
        /// </summary>
        /// <returns>printerJobID</returns>
        static int CheckSupplies(BidiSplWrap bidiSpl, PrinterOptionsValues printerOptionsValues)
        {
            int checkPrintRibbon = (printerOptionsValues._printHead == "Installed") ? 1 : 0;
            int checkEmbossSupplies = printerOptionsValues._moduleEmbosser == "None" ? 0 : 1;

            string startJobXml = string.Format(
                strings.STARTJOB_XML,
                "",     // use default input hopper in this sample
                checkPrintRibbon,
                checkEmbossSupplies);

            string printerStatusXML = bidiSpl.SetPrinterData(strings.STARTJOB, startJobXml);
            PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
            if (0 != printerStatusValues._errorCode)
            {
                // the StartJob failed; send a Cancel(); do NOT send an EndJob():
                CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode);
                throw new Exception("Canceling the job. Printer Error: " + printerStatusValues._errorString);
            }

            return printerStatusValues._printerJobID;
        }

        static void Main(string[] args)
        {
            CommandLineOptions commandLineOptions = CommandLineOptions.CreateFromArguments(args);
            commandLineOptions.Validate();

            BidiSplWrap bidiSpl = null;
            try
            {
                bidiSpl = new BidiSplWrap();
                bidiSpl.BindDevice(commandLineOptions.printerName);

                string sdkVersion = bidiSpl.GetPrinterData(strings.SDK_VERSION);
                Console.WriteLine(Environment.NewLine + "SDK Version: " + Util.ParseSDKVersionXML(sdkVersion) + Environment.NewLine);

                string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);
                PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);
                if ("Ready" != printerOptionsValues._printerStatus)
                {
                    throw new Exception(commandLineOptions.printerName + " is not ready.");
                }

                int printerJobID = 0;
                if (commandLineOptions.checkSupplies)
                {
                    printerJobID = CheckSupplies(bidiSpl, printerOptionsValues);
                }

                SamplePrintDocument printDocument = new SamplePrintDocument(commandLineOptions);
                printDocument.PrintController = new StandardPrintController(); // suppress annoying progress popups
                printDocument.BeginPrint += new PrintEventHandler(printDocument.OnBeginPrint);
                printDocument.QueryPageSettings += new QueryPageSettingsEventHandler(printDocument.OnQueryPageSettings);
                printDocument.PrintPage += new PrintPageEventHandler(printDocument.OnPrintPage);
                printDocument.Print();

                if (0 != printerJobID)
                {
                    // we did a StartJob() to check for supplies. So, wait a
                    // moment for the spooling to finish, and then issue an EndJob():
                    Util.WaitForWindowsJobID(bidiSpl, commandLineOptions.printerName);
                    bidiSpl.SetPrinterData(strings.ENDJOB);
                }
            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.Message);
            }
            finally
            {
                bidiSpl.UnbindDevice();
            }
        }
    }
}
