﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// xps card printer SDK: smartcard sample
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Threading;
using System.Windows.Forms;
using dxp01sdk;

public class CommandLineOptions
{
    public string printerName;
    public bool print;
    public bool pollForJobCompletion;
};

class smartcard
{
    private static void Usage()
    {
        Console.WriteLine("Smartcard.exe demonstrates interactive mode parking a card in the smart card");
        Console.WriteLine("  station, moving the card from the station, and options to print and poll for");
        Console.WriteLine("  job completion.");
        Console.WriteLine();
        Console.WriteLine("> Uses hard-coded data for printing.");
        Console.WriteLine();
        Console.WriteLine("smartcard.exe -n <printername> [-p] [-c]");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine("  -p Print text and graphic on front side of card.");
        Console.WriteLine("  -c Poll for job status.");
        Console.WriteLine();
        Console.WriteLine("smartcard.exe -n \"XPS Card Printer\"");
        Console.WriteLine("  Parks a card in the printer smart card station, asks you to continue or");
        Console.WriteLine("  reject and then does what you requested.");
        Console.WriteLine();
        Console.WriteLine("smartcard.exe -n \"XPS Card Printer\" -p -c");
        Console.WriteLine("  Parks a card in printer smart card station, moves the card from the station,");
        Console.WriteLine("  prints simple text and graphics on the front side of the card, and polls and");
        Console.WriteLine("  displays job status.");
        Environment.Exit(-1);
    }

    static CommandLineOptions GetCommandlineOptions(string[] args)
    {
        if (args.Length == 0) Usage();

        CommandLineOptions commandLineOptions = new CommandLineOptions();
        CommandLine.Utility.Arguments arguments = new CommandLine.Utility.Arguments(args);

        if (string.IsNullOrEmpty(arguments["n"]))
        {
            Console.WriteLine("printer name is required");
            Environment.Exit(-1);
        }
        else
        {
            // we might have a -n with no printer name:
            bool boolVal = false;
            if (Boolean.TryParse(arguments["n"], out boolVal))
            {
                Console.WriteLine("printer name is required");
                Environment.Exit(-1);
            }
            commandLineOptions.printerName = arguments["n"];
        }

        commandLineOptions.print = !string.IsNullOrEmpty(arguments["p"]);
        commandLineOptions.pollForJobCompletion = !string.IsNullOrEmpty(arguments["c"]);

        return commandLineOptions;
    }

    static void ResumeJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
    {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Resume, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode)
    {
        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);
    }

    /// <summary>
    /// Simulate a smartcard personalization operation.
    /// 
    /// This is where calls to a third-party library would be used to communicate
    /// with the chip.
    /// </summary>
    /// <returns></returns>
    static bool PersonalizeSmartcard()
    {
        string caption = "A card has been parked in smartcard module";
        string message =
           "This is where smartcard personalization happens.\r\n\r\n" +
           "Press 'Yes' to simulate a successful smartcard personalization.\r\n\r\n" +
           "Press 'No' to simulate a failed smartcard personalization.";

        DialogResult result = MessageBox.Show(message, caption, MessageBoxButtons.YesNo);
        return (result == DialogResult.Yes);
    }

    static void PollForJobCompletion(BidiSplWrap bidiSpl, int printerJobID)
    {
        for (; true; )
        {
            string temp = strings.JOB_STATUS_XML;
            string jobInfoXML = string.Format(temp, printerJobID);
            string jobStatusXML = bidiSpl.GetPrinterData(strings.JOB_STATUS, jobInfoXML);
            JobStatusValues jobStatusValues = Util.ParseJobStatusXML(jobStatusXML);

            Console.WriteLine("Printer job id:" + printerJobID + " " + jobStatusValues._jobState);

            if (strings.JOB_ACTIVE != jobStatusValues._jobState)
            {
                return;
            }
            Thread.Sleep(1000);
        }
    }

    static void StartJob(BidiSplWrap bidiSpl, ref int printerJobID)
    {
        string printerStatusXML = bidiSpl.SetPrinterData(strings.STARTJOB);
        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("StarJob() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);
            CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode);
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }
        else
        {
            printerJobID = printerStatusValues._printerJobID;
        }
    }

    static void SmartcardPark(BidiSplWrap bidiSpl)
    {
        string printerStatusXML = bidiSpl.SetPrinterData(strings.SMARTCARD_PARK);
        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("SmartcardPark() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);
            CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode);
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }
    }

    static void SmartcardPersonalizationFailed(BidiSplWrap bidiSpl, int printerJobID)
    {
        Console.WriteLine("SmartcardPersonalizationFailed(); canceling the job.");
        CancelJob(bidiSpl, printerJobID, 0);
        bidiSpl.UnbindDevice();
        Environment.Exit(-1);
    }

    ////////////////////////////////////////////////////////////////////////////////
    // Main()
    ////////////////////////////////////////////////////////////////////////////////
    static void Main(string[] args)
    {
        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = new BidiSplWrap();
        bidiSpl.BindDevice(commandLineOptions.printerName);

        // Get SDK version 
        string sdkVersion = bidiSpl.GetPrinterData(strings.SDK_VERSION);
        Console.WriteLine("\nSDK Version: " + Util.ParseSDKVersionXML(sdkVersion) + "\n");

        // is printer is ready to accept a smartcard job?
        string printerOptionsXML = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);

        PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXML);
        if ("Ready" != printerOptionsValues._printerStatus)
        {
            Console.WriteLine(commandLineOptions.printerName + " is not ready.");
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        // start an IBidiSpl job:
        int printerJobID = 0;
        StartJob(bidiSpl, ref printerJobID);

        // Park the card at smartcard station:
        SmartcardPark(bidiSpl);

        // The card has been parked in the smartcard station. Personalize the smart card.
        bool personalizedOK = PersonalizeSmartcard();
        if (!personalizedOK)
        {
            SmartcardPersonalizationFailed(bidiSpl, printerJobID);
        }

        // The Smartcard Personalization succeeded; send Resume command
        ResumeJob(bidiSpl, printerJobID, 0);

        if (commandLineOptions.print)
        {
            Util.PrintTextAndGraphics(bidiSpl, commandLineOptions.printerName);
        }

        // End the smartcard job
        bidiSpl.SetPrinterData(strings.ENDJOB);

        if (commandLineOptions.pollForJobCompletion)
        {
            PollForJobCompletion(bidiSpl, printerJobID);
        }

        bidiSpl.UnbindDevice();
    }
}