﻿////////////////////////////////////////////////////////////////////////////////
// Copyright (c) Datacard Corporation.  All Rights Reserved.
//
// XPS Printer Driver SDK: Single-wire smartcard personalization c# sample.
////////////////////////////////////////////////////////////////////////////////
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
using dxp01sdk;

public class CommandLineOptions
{
    public string printerName;
    public bool print;
    public bool pollForJobCompletion;
    public bool verbose;
};

class smartcard_singlewire
{
    private static void Usage()
    {
        string thisExeName = Path.GetFileName(Environment.GetCommandLineArgs()[0]);
        Console.Write(thisExeName);
        Console.WriteLine(" demonstrates interactive mode parking a card in the");
        Console.WriteLine("  smart card station, performing single-wire smartcard functions, moving");
        Console.WriteLine("  the card from the station, and options to print and poll for job completion.");
        Console.WriteLine();
        Console.WriteLine("  Uses hard-coded data for printing.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n <printername> [-p] [-c] [-v]");
        Console.WriteLine();
        Console.WriteLine("options:");
        Console.WriteLine("  -n <printername>. Required. Try -n \"XPS Card Printer\".");
        Console.WriteLine("  -p Print text and graphic on front side of card.");
        Console.WriteLine("  -c Poll for job status.");
        Console.WriteLine("  -v Display detailed information with timings.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n \"XPS Card Printer\"");
        Console.WriteLine("  Parks a card in the printer smart card station, asks you to continue or");
        Console.WriteLine("  reject and then does what you requested.");
        Console.WriteLine();
        Console.WriteLine(thisExeName + " -n \"XPS Card Printer\" -p -c");
        Console.WriteLine("  Parks a card in printer smart card station, moves the card from the station,");
        Console.WriteLine("  prints simple text and graphics on the front side of the card, and polls and");
        Console.WriteLine("  displays job status.");
        Environment.Exit(-1);
    }

    static CommandLineOptions GetCommandlineOptions(string[] args)
    {
        if (args.Length == 0) Usage();

        var commandLineOptions = new CommandLineOptions();
        var arguments = new CommandLine.Utility.Arguments(args);

        if (string.IsNullOrEmpty(arguments["n"]))
        {
            Console.WriteLine("printer name is required");
            Environment.Exit(-1);
        }
        else
        {
            // we might have a -n with no printer name:
            bool boolVal = false;
            if (Boolean.TryParse(arguments["n"], out boolVal))
            {
                Console.WriteLine("printer name is required");
                Environment.Exit(-1);
            }
            commandLineOptions.printerName = arguments["n"];
        }

        commandLineOptions.print = !string.IsNullOrEmpty(arguments["p"]);
        commandLineOptions.pollForJobCompletion = !string.IsNullOrEmpty(arguments["c"]);
        commandLineOptions.verbose = !string.IsNullOrEmpty(arguments["v"]);

        return commandLineOptions;
    }

    static void ResumeJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " ResumeJob() start"); }

        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Resume, printerJobID, errorCode);
        Console.WriteLine("issuing Resume after smartcard:");
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " ResumeJob() return"); }
    }

    static void CancelJob(BidiSplWrap bidiSpl, int printerJobID, int errorCode, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " CancdlJob() start"); }

        string xmlFormat = strings.PRINTER_ACTION_XML;
        string input = string.Format(xmlFormat, (int)Actions.Cancel, printerJobID, errorCode);
        bidiSpl.SetPrinterData(strings.PRINTER_ACTION, input);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " CancdlJob() return"); }
    }

    ////////////////////////////////////////////////////////////////////////////////
    // SCardResultMessage()
    // format a message for display.
    // SCARD errors are declared in WinError.H
    ////////////////////////////////////////////////////////////////////////////////
    static string SCardResultMessage(long scardResult, string message)
    {
        var sb = new StringBuilder();
        sb.Append(message);
        sb.Append(" result: ");
        sb.Append(scardResult.ToString());
        sb.Append("; ");
        sb.Append(Util.Win32ErrorString((int)scardResult));
        return sb.ToString();
    }

    ////////////////////////////////////////////////////////////////////////////
    /// <summary>
    /// a card is in the printer and stated in the smartcard module. The
    /// SCardConnect() call has succeeded.
    /// talk to the reader and the chip (contact or contactless).
    /// </summary>
    /// <param name="scard">an instance of our SCard wrapper class.</param>
    ////////////////////////////////////////////////////////////////////////////
    static void PersonalizeConnectedChip(SCard scard, bool verbose)
    {
        int[] states = new int[0];
        uint protocol = (uint)scard_protocol.SCARD_PROTOCOL_UNDEFINED;
        byte[] ATRBytes = new byte[0];

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeConnectedChip() calling SCardStatus()"); }

        var scardResult = scard.SCardStatus(ref states, ref protocol, ref ATRBytes);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardStatus() return"); }

        Console.WriteLine("SCardStatus result: {0}: {1}", scardResult, Util.Win32ErrorString((int)scardResult));
        if (scard_error.SCARD_S_SUCCESS != (scard_error)scardResult)
        {
            // unable to get a valid SCardStatus() response; we're done.
            return;
        }

        bool cardAbsent = false;
        Console.Write("SCardStatus() states: ");
        for (var index = 0; index < states.Length; index++)
        {
            Console.Write(states[index]);
            Console.Write(" ");
            Console.Write(SCard.StringFromState(states[index]));
            Console.Write(" ");

            if (scard_state.SCARD_ABSENT == (scard_state)states[index])
                cardAbsent = true;
        }
        Console.WriteLine();

        if (cardAbsent)
        {
            // one of the states is SCARD_ABSENT; we're done.
            return;
        }

        Console.Write("SCardStatus() ATRBytes (hex): ");
        for (var index = 0; index < ATRBytes.Length; index++)
        {
            Console.Write("{0:X02} ", ATRBytes[index]);
        }
        Console.WriteLine();

        var attributesToTry = new scard_attr[] {
            scard_attr.SCARD_ATTR_VENDOR_NAME,
            scard_attr.SCARD_ATTR_VENDOR_IFD_SERIAL_NO,
            scard_attr.SCARD_ATTR_VENDOR_IFD_TYPE,
            scard_attr.SCARD_ATTR_VENDOR_IFD_VERSION
        };

        for (var attribIndex = 0; attribIndex < attributesToTry.Length; attribIndex++)
        {
            byte[] scardAttributeBytes = new byte[0];

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeConnectedChip() calling SCardGetAttrib()"); }

            scardResult = scard.SCardGetAttrib(attributesToTry[attribIndex], ref scardAttributeBytes);

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardGetAttrib() return"); }

            if (scard_error.SCARD_S_SUCCESS != (scard_error)scardResult)
                continue;

            string msg = string.Format("SCardGetAttrib({0})", SCard.StringFromAttr((int)attributesToTry[attribIndex]));
            Console.WriteLine(SCardResultMessage(scardResult, msg));

            switch (attributesToTry[attribIndex])
            {
                case scard_attr.SCARD_ATTR_VENDOR_IFD_VERSION:
                    // vendor-supplied interface device version: DWORD in the form 0xMMmmbbbb where
                    //      MM = major version;
                    //      mm = minor version; and
                    //      bbbb = build number:
                    var byteCount = scardAttributeBytes.Length;
                    var minorVersion = byteCount > 0 ? scardAttributeBytes[0] : 0;
                    var majorVersion = byteCount > 1 ? scardAttributeBytes[1] : 0;
                    var buildNumber = 0;
                    if (byteCount > 3)
                    {
                        buildNumber = (scardAttributeBytes[2] << 8) + scardAttributeBytes[3];
                    }

                    Console.WriteLine("  SCARD_ATTR_VENDOR_IFD_VERSION:");
                    Console.WriteLine("    major: " + majorVersion.ToString());
                    Console.WriteLine("    minor: " + minorVersion.ToString());
                    Console.WriteLine("    build: " + buildNumber.ToString());
                    break;

                default:
                    // null-terminate the returned byte for string display.
                    var bytesList = new List<byte>(scardAttributeBytes);
                    bytesList.Add(0);
                    scardAttributeBytes = bytesList.ToArray();
                    Console.WriteLine(ASCIIEncoding.UTF8.GetString(scardAttributeBytes));
                    break;
            }
        }

        // create a bytes for the upcoming SCardTransmit() method.
        // these particular bytes should function with this type of contact chip:
        //
        //      MPCOS-EMV 16k
        //      GEMPLUS
        //      Datacard part number 806062-002
        //
        byte[] sendBytes = new byte[] { 0x00, 0xA4, 0x00, 0x00 };
        Console.Write("sending bytes : ");
        for (var index = 0; index < sendBytes.Length; index++)
        {
            Console.Write("{0:X02} ", sendBytes[index]);
        }
        Console.WriteLine();

        // for those bytes, we should receive 0x61, 0x12
        byte[] receivedBytes = new byte[0];

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeConnectedChip() calling SCardTransmit()"); }

        scardResult = scard.SCardTransmit(sendBytes, ref receivedBytes);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardTransmit() return"); }

        Console.WriteLine("SCardTransmit result: {0}", scardResult);

        if (scard_error.SCARD_S_SUCCESS == (scard_error)scardResult && receivedBytes.Length > 0)
        {
            Console.Write("  {0} bytes received: ", receivedBytes.Length);
            for (var index = 0; index < receivedBytes.Length; index++)
            {
                Console.Write("{0:X02} ", receivedBytes[index]);
            }
            Console.WriteLine();
        }

        // send the bytes       0x00  0xC0  0x00  0x00  0x12:
        sendBytes = new byte[] { 0x00, 0xC0, 0x00, 0x00, 0x12 };
        Console.Write("sending bytes : ");
        for (var index = 0; index < sendBytes.Length; index++)
        {
            Console.Write("{0:X02} ", sendBytes[index]);
        }
        Console.WriteLine();

        // for those bytes, we should receive
        //  0x85, 0x10, 0x80, 0x01, 0x3F, 0x00, 0x38, 0x00, 0x00, 0x00,
        //  0x01, 0x00, 0x01, 0x00, 0x00, 0x00, 0x00, 0x6B, 0x90, 0x00
        receivedBytes = new byte[0];

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeConnectedChip() calling SCardTransmit()"); }

        scardResult = scard.SCardTransmit(sendBytes, ref receivedBytes);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeConnectedChip() SCardTransmit() return"); }

        Console.WriteLine("SCardTransmit result: {0}", scardResult);

        if (scard_error.SCARD_S_SUCCESS == (scard_error)scardResult && receivedBytes.Length > 0)
        {
            Console.Write("  {0} bytes received: ", receivedBytes.Length);
            for (var index = 0; index < receivedBytes.Length; index++)
            {
                Console.Write("{0:X02} ", receivedBytes[index]);
            }
            Console.WriteLine();
        }
    }

    /// <summary>
    /// Demonstrate the single-wire smartcard calls. The wrapper class we consume
    /// is similar to the Win32 ::Scard...() functions.
    /// <returns>
    /// returns FALSE if either the contact or contactless personalization failed.
    /// for this sample, we use the result of the SCardConnect() for pass/fail.
    /// </returns>
    /// </summary>
    static bool PersonalizeSmartcard(BidiSplWrap bidiSpl, bool verbose)
    {
        var scard = new SCard(bidiSpl);
        var connectedContactOK = false;

        // try a 'contact' chip:
        uint protocol = (uint)scard_protocol.SCARD_PROTOCOL_DEFAULT;

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeSmartcard() calling SCardConnect()"); }

        long scardResult = scard.SCardConnect(SCard.ChipConnection.contact, ref protocol);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardConnect() return"); }

        Console.WriteLine("SCardConnect(contact) result: {0}", scardResult);
        if (scard_error.SCARD_S_SUCCESS == (scard_error)scardResult)
        {
            connectedContactOK = true;
            var protocols = SCard.StringsFromProtocol(protocol);
            Console.Write("   protocol[s]: ");
            for (int index = 0; index < protocols.Length; index++)
            {
                Console.Write(protocols[index] + " ");
            }
            Console.WriteLine();

            PersonalizeConnectedChip(scard, verbose);

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeSmartcard() calling SCardDisConnect()"); }

            scardResult = scard.SCardDisConnect((int)scard_disposition.SCARD_LEAVE_CARD);

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardDisConnect() return"); }

            Console.WriteLine("SCardDisConnect() result: {0}", scardResult);
        }

        // try a 'contactless' chip:
        var connectedContactlessOK = false;
        protocol = (uint)scard_protocol.SCARD_PROTOCOL_DEFAULT;

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeSmartcard() calling SCardConnect()"); }

        scardResult = scard.SCardConnect(SCard.ChipConnection.contactless, ref protocol);

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardConnect() return"); }

        Console.WriteLine("SCardConnect(contactless) result: {0}", scardResult);
        if (scard_error.SCARD_S_SUCCESS == (scard_error)scardResult)
        {
            connectedContactlessOK = true;
            var protocols = SCard.StringsFromProtocol(protocol);
            Console.Write("   protocol[s]: ");
            for (int index = 0; index < protocols.Length; index++)
            {
                Console.Write(protocols[index] + " ");
            }
            Console.WriteLine();

            PersonalizeConnectedChip(scard, verbose);

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PersonalizeSmartcard() calling SCardDisConnect()"); }

            scardResult = scard.SCardDisConnect((int)scard_disposition.SCARD_LEAVE_CARD);

            if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SCardDisConnect() return"); }

            Console.WriteLine("SCardDisConnect() result: {0}", scardResult);
        }

        return connectedContactOK && connectedContactlessOK;
    }

    static void PollForJobCompletion(BidiSplWrap bidiSpl, int printerJobID, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PollForJobCompletion() start"); }

        for (; true; )
        {

            string printerStatusXML = bidiSpl.GetPrinterData(strings.PRINTER_MESSAGES);
            PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
            if (0 != printerStatusValues._errorCode)
            {
                Console.WriteLine(printerStatusValues._errorString + " severity: " + printerStatusValues._errorSeverity.ToString());
                Console.WriteLine("canceling printer jobID " + printerStatusValues._printerJobID.ToString());
                CancelJob(
                    bidiSpl,
                    printerStatusValues._printerJobID,
                    printerStatusValues._errorCode,
                    verbose);
                return;
            }

            string xmlFormat = strings.JOB_STATUS_XML;
            string jobInfoXML = string.Format(xmlFormat, printerJobID);
            string jobStatusXML = bidiSpl.GetPrinterData(strings.JOB_STATUS, jobInfoXML);
            JobStatusValues jobStatusValues = Util.ParseJobStatusXML(jobStatusXML);

            Console.WriteLine("Printer job id: " + printerJobID + " " + jobStatusValues._jobState);

            if (strings.JOB_ACTIVE != jobStatusValues._jobState)
            {
                return;
            }
            Thread.Sleep(1000);
        }
    }

    static void StartJob(BidiSplWrap bidiSpl, ref int printerJobID, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " StartJob() start"); }

        string startJobXml = string.Format(
            strings.STARTJOB_XML,
            "",     // use default input hopper
            0,      // zero means "do not check print ribbon"
            0);     // zero means "do not check emboss supplies"

        string printerStatusXML = bidiSpl.SetPrinterData(strings.STARTJOB, startJobXml);

        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("StartJob() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);
            CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode, verbose);
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }
        else
        {
            printerJobID = printerStatusValues._printerJobID;
        }

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " StartJob() return"); }
    }

    static void SmartcardPark(BidiSplWrap bidiSpl, bool verbose)
    {
        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SmartcardPark() start"); }

        string printerStatusXML = bidiSpl.SetPrinterData(strings.SMARTCARD_PARK);
        PrinterStatusValues printerStatusValues = Util.ParsePrinterStatusXML(printerStatusXML);
        if (0 != printerStatusValues._errorCode)
        {
            Console.WriteLine("SmartcardPark() fail; canceling the job. Printer Error: " + printerStatusValues._errorString);
            CancelJob(bidiSpl, printerStatusValues._printerJobID, printerStatusValues._errorCode, verbose);
            bidiSpl.UnbindDevice();
            Environment.Exit(-1);
        }

        if (verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " SmartcardPark() return"); }
    }

    static void SmartcardPersonalizationFailed(BidiSplWrap bidiSpl, int printerJobID, bool verbose)
    {
        Console.WriteLine("SmartcardPersonalizationFailed(); canceling the job.");
        CancelJob(bidiSpl, printerJobID, 0, verbose);
        Environment.Exit(-1);
    }

    ////////////////////////////////////////////////////////////////////////////
    // Main()
    ////////////////////////////////////////////////////////////////////////////
    static void Main(string[] args)
    {
        CommandLineOptions commandLineOptions = GetCommandlineOptions(args);

        BidiSplWrap bidiSpl = new BidiSplWrap();
        bidiSpl.BindDevice(commandLineOptions.printerName);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling GetPrinterData(version)"); }

        string sdkVersionXml = bidiSpl.GetPrinterData(strings.SDK_VERSION);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " GetPrinterData(version) return"); }

        Console.WriteLine(Environment.NewLine + "SDK Version: " + Util.ParseSDKVersionXML(sdkVersionXml) + Environment.NewLine);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling GetPrinterData(options)"); }

        // see if the printer is in the Printer_Ready state:
        string printerOptionsXml = bidiSpl.GetPrinterData(strings.PRINTER_OPTIONS2);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " GetPrinterData(options) return"); }

        PrinterOptionsValues printerOptionsValues = Util.ParsePrinterOptionsXML(printerOptionsXml);
        if ("Ready" != printerOptionsValues._printerStatus)
        {
            Console.WriteLine(commandLineOptions.printerName + " is not ready.");
            Environment.Exit(-1);
        }

        // this application uses the single-wire method for communicating with
        // a smartcard reader in the printer. If the printer is not configured with
        // that option, we're done.
        if ("Single wire" != printerOptionsValues._optionSmartcard)
        {
            Console.WriteLine("printer '" + commandLineOptions.printerName + "' does not have single-wire smartcard option. Exiting.");
            Environment.Exit(-1);
        }

        // start an IBidiSpl job:
        int printerJobID = 0;
        StartJob(bidiSpl, ref printerJobID, commandLineOptions.verbose);

        // Park the card at smartcard station:
        SmartcardPark(bidiSpl, commandLineOptions.verbose);

        // The card has been parked in the smartcard station. Personalize the smart card.
        bool personalizedOK = false;
        try
        {
            personalizedOK = PersonalizeSmartcard(bidiSpl, commandLineOptions.verbose);
        }
        catch (Exception e)
        {
            Console.WriteLine(e.Message);
        }

        if (!personalizedOK)
        {
            SmartcardPersonalizationFailed(bidiSpl, printerJobID, commandLineOptions.verbose);
        }

        // The Smartcard Personalization succeeded; send Resume command
        ResumeJob(bidiSpl, printerJobID, 0, commandLineOptions.verbose);

        if (commandLineOptions.print)
        {
            if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling PrintTextAndGraphics()"); }
            Util.PrintTextAndGraphics(bidiSpl, commandLineOptions.printerName);
            if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " PrintTextAndGraphics() return"); }
        }

        // End the smartcard job
        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " calling endjob"); }

        bidiSpl.SetPrinterData(strings.ENDJOB);

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " return from endjob"); }

        if (commandLineOptions.pollForJobCompletion)
        {
            PollForJobCompletion(bidiSpl, printerJobID, commandLineOptions.verbose);
        }

        if (commandLineOptions.verbose) { Console.WriteLine(Util.GetCurrentTimeString() + " end."); }

        bidiSpl.UnbindDevice();
    }
}
